# sage_setup: distribution = sagemath-repl
"""
The Sage ZMQ Kernel

Version of the Jupyter kernel when running Sage inside the Jupyter
notebook or remote Jupyter sessions.
"""

# ***************************************************************************
#       Copyright (C) 2015 Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

import sys
import warnings
with warnings.catch_warnings():
    # When upstream pydevd (as opposed to the bundled version) is used
    # with debugpy, a PEP 420 warning is emitted. Debugpy and/or
    # pydevd will eventually work around this, but as of September
    # 2023, hiding the warning gives us more flexibility in the
    # versions of those packages that we can accept.
    warnings.filterwarnings("ignore",
                            message=r".*pkg_resources\.declare_namespace",
                            category=DeprecationWarning)
    from ipykernel.ipkernel import IPythonKernel

from ipykernel.zmqshell import ZMQInteractiveShell
from traitlets import Type

from sage.env import SAGE_VERSION
from sage.repl.interpreter import SageNotebookInteractiveShell
from sage.repl.ipython_extension import SageJupyterCustomizations


class SageZMQInteractiveShell(SageNotebookInteractiveShell, ZMQInteractiveShell):
    pass


class SageKernel(IPythonKernel):
    implementation = 'sage'
    implementation_version = SAGE_VERSION

    shell_class = Type(SageZMQInteractiveShell)

    def __init__(self, **kwds):
        """
        The Sage Jupyter Kernel.

        INPUT:

        See the Jupyter documentation.

        EXAMPLES::

            sage: from sage.repl.ipython_kernel.kernel import SageKernel
            sage: SageKernel.__new__(SageKernel)
            <sage.repl.ipython_kernel.kernel.SageKernel object at 0x...>
        """
        super().__init__(**kwds)
        SageJupyterCustomizations(self.shell)

    @property
    def banner(self):
        r"""
        The Sage Banner.

        The value of this property is displayed in the Jupyter
        notebook.

        OUTPUT: string

        EXAMPLES::

            sage: from sage.repl.ipython_kernel.kernel import SageKernel
            sage: sk = SageKernel.__new__(SageKernel)
            sage: print(sk.banner)
            ┌...SageMath version...
        """
        from sage.misc.banner import banner_text
        return banner_text()

    @property
    def help_links(self):
        r"""
        Help in the Jupyter Notebook.

        OUTPUT: see the Jupyter documentation

        EXAMPLES::

            sage: from sage.repl.ipython_kernel.kernel import SageKernel
            sage: sk = SageKernel.__new__(SageKernel)
            sage: sk.help_links
            [{'text': 'Sage Documentation',
              'url': '.../html/en/index.html'},
             ...]
        """
        # A Sage doc server starts when Jupyter notebook launches if the Sage
        # documentation is available locally.  See the corresponding code in
        # src/bin/sage-notebook.

        from sage.env import SAGE_DOC_SERVER_URL
        from sage.env import SAGE_DOC_LOCAL_PORT as port
        from sage.features.sagemath import sagemath_doc_html

        if SAGE_DOC_SERVER_URL:
            def doc_url(path):
                return f'{SAGE_DOC_SERVER_URL}/{path}'
        elif sagemath_doc_html().is_present() and int(port):
            def doc_url(path):
                return f'http://127.0.0.1:{port}/{path}'
        else:
            def doc_url(path):
                return f'https://doc.sagemath.org/{path}'

        return [
            {
                'text': 'Sage Documentation',
                'url': doc_url('html/en/index.html'),
            },
            {
                'text': 'A Tour of Sage',
                'url': doc_url('html/en/a_tour_of_sage/index.html'),
            },
            {
                'text': 'Tutorial',
                'url': doc_url('html/en/tutorial/index.html'),
            },
            {
                'text': 'Thematic Tutorials',
                'url': doc_url('html/en/thematic_tutorials/index.html'),
            },
            {
                'text': 'PREP Tutorials',
                'url': doc_url('html/en/prep/index.html'),
            },
            {
                'text': 'Constructions',
                'url': doc_url('html/en/constructions/index.html'),
            },
            {
                'text': 'FAQ',
                'url': doc_url('html/en/faq/index.html'),
            },
            {
                'text': 'Reference Manual',
                'url': doc_url('html/en/reference/index.html'),
            },
            {
                'text': "Installation Guide",
                'url': doc_url('html/en/installation/index.html'),
            },
            {
                'text': "Developer Guide",
                'url': doc_url('html/en/developer/index.html'),
            },
            {
                'text': "Python",
                'url': "http://docs.python.org/%i.%i" % sys.version_info[:2],
            },
            {
                'text': "IPython",
                'url': "http://ipython.org/documentation.html",
            },
            {
                'text': 'Singular',
                'url': 'http://www.singular.uni-kl.de/Manual/latest/index.htm',
            },
            {
                'text': 'GAP',
                'url': 'http://gap-system.org/Manuals/doc/ref/chap0.html',
            },
            {
                'text': "NumPy",
                'url': "http://docs.scipy.org/doc/numpy/reference/",
            },
            {
                'text': "SciPy",
                'url': "http://docs.scipy.org/doc/scipy/reference/",
            },
            {
                'text': "SymPy",
                'url': 'http://docs.sympy.org/latest/index.html',
            },
            {
                'text': "Matplotlib",
                'url': "https://matplotlib.org/contents.html",
            },
            {
                'text': "Markdown",
                'url': "http://help.github.com/articles/github-flavored-markdown",
            },
        ]

    def pre_handler_hook(self):
        """
        Restore the signal handlers to their default values at Sage
        startup, saving the old handler at the ``saved_sigint_handler``
        attribute. This is needed because Jupyter needs to change the
        ``SIGINT`` handler.

        See :issue:`19135`.

        TESTS::

            sage: from sage.repl.ipython_kernel.kernel import SageKernel
            sage: k = SageKernel.__new__(SageKernel)
            sage: k.pre_handler_hook()
            sage: k.saved_sigint_handler
            <cyfunction python_check_interrupt at ...>
        """
        from cysignals import init_cysignals
        self.saved_sigint_handler = init_cysignals()



# sage_setup: distribution = sagemath-repl
r"""
Sage's IPython Modifications

This module contains all of Sage's customizations to the IPython
interpreter.  These changes consist of the following major components:

  - :class:`SageTerminalApp`
  - :class:`SageInteractiveShell`
  - :class:`SageTerminalInteractiveShell`
  - :func:`interface_shell_embed`

SageTerminalApp
---------------

This is the main application object.  It is used by the
``$SAGE_LOCAL/bin/sage-ipython`` script to start the Sage
command-line.  It's primary purpose is to

  - Initialize the :class:`SageTerminalInteractiveShell`.

  - Provide default configuration options for the shell, and its
    subcomponents.  These work with (and can be overridden by)
    IPython's configuration system.

  - Load the Sage ipython extension (which does things like preparsing,
    add magics, etc.).

  - Provide a custom :class:`SageCrashHandler` to give the user
    instructions on how to report the crash to the Sage support
    mailing list.

SageInteractiveShell
--------------------

The :class:`SageInteractiveShell` object is the object responsible for
accepting input from the user and evaluating it.  From the command-line,
this object can be retrieved by running::

    sage: shell = get_ipython()   # not tested

Any input is preprocessed and evaluated inside the ``shell.run_cell``
method. If the command line processing does not do what you want it to
do, you can step through it in the debugger::

    sage: %debug shell.run_cell('?')        # not tested

The :class:`SageInteractiveShell` provides the following
customizations:

  - Modify the libraries before calling system commands. See
    :meth:`~SageInteractiveShell.system_raw`.

SageTerminalInteractiveShell
----------------------------

The :class:`SageTerminalInteractiveShell` is a close relative of
:class:`SageInteractiveShell` that is specialized for running in a
terminal. In particular, running commands like ``!ls`` will directly
write to stdout. Technically, the ``system`` attribute will point to
``system_raw`` instead of ``system_piped``.

Interface Shell
---------------

The function :func:`interface_shell_embed` takes a
:class:`~sage.interfaces.interface.Interface` object and returns an
embeddable IPython shell which can be used to directly interact with
that shell.  The bulk of this functionality is provided through
:class:`InterfaceShellTransformer`.

TESTS:

Check that Cython source code appears in tracebacks::

    sage: from sage.repl.interpreter import get_test_shell
    sage: shell = get_test_shell()
    sage: shell.run_cell('1/0') # known bug (meson doesn't include the Cython source code)
    ...
    ZeroDivisionError...Traceback (most recent call last)
    ...
    ----> 1 Integer(1)/Integer(0)
    ...
    ...integer.pyx... in sage.rings.integer.Integer...div...
    ...
    -> ...                  raise ZeroDivisionError("rational division by zero")
       ....:            x = <Rational> Rational.__new__(Rational)
       ....:            mpq_div_zz(x.value, ....value, (<Integer>right).value)
    <BLANKLINE>
    ZeroDivisionError: rational division by zero
    sage: shell.quit()

Test prompt transformer::

    sage: from sage.repl.interpreter import SagePromptTransformer
    sage: spt = SagePromptTransformer
    sage: spt(["sage: 2 + 2"])
    ['2 + 2']
    sage: spt([''])
    ['']
    sage: spt(["....: 2+2"])
    ['2+2']

This should strip multiple prompts: see :issue:`16297`::

    sage: spt(["sage:   sage: 2+2"])
    ['2+2']
    sage: spt(["   sage: ....: 2+2"])
    ['2+2']

The prompt contains a trailing space. Extra spaces between the
last prompt and the remainder should not be stripped::

    sage: spt(["   sage: ....:    2+2"])
    ['   2+2']

We test that the input transformer is enabled on the Sage command
line::

    sage: from sage.repl.interpreter import get_test_shell
    sage: shell = get_test_shell()
    sage: shell.run_cell('sage: a = 123')              # single line
    sage: shell.run_cell('sage: a = [\n... 123]')      # old-style multi-line
    sage: shell.run_cell('sage: a = [\n....: 123]')    # new-style multi-line

We test that :issue:`16196` is resolved::

    sage: shell.run_cell('    sage: 1+1')
    2
    sage: shell.quit()
"""

# ****************************************************************************
#       Copyright (C) 2004-2012 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import re
from traitlets import Bool, Type

from sage.repl.preparse import preparse, containing_block
from sage.repl.prompts import InterfacePrompts
from sage.repl.configuration import sage_ipython_config, SAGE_EXTENSION

from IPython.core.interactiveshell import InteractiveShell
from IPython.terminal.interactiveshell import TerminalInteractiveShell
from IPython.core.inputtransformer2 import PromptStripper
from IPython.core.prefilter import PrefilterTransformer
from IPython.terminal.embed import InteractiveShellEmbed
from IPython.terminal.ipapp import TerminalIPythonApp, IPAppCrashHandler
from IPython.core.crashhandler import CrashHandler

from ctypes import pythonapi, c_int, c_void_p
# The following functions are part of the stable ABI since python 3.2
# See: https://docs.python.org/3/c-api/sys.html#c.PyOS_getsig

# PyOS_sighandler_t PyOS_getsig(int i)
pythonapi.PyOS_getsig.restype = c_void_p
pythonapi.PyOS_getsig.argtypes = c_int,

# PyOS_sighandler_t PyOS_setsig(int i, PyOS_sighandler_t h)
pythonapi.PyOS_setsig.restype = c_void_p
pythonapi.PyOS_setsig.argtypes = c_int, c_void_p,


# TODO: This global variable _do_preparse should be associated with an
# IPython InteractiveShell as opposed to a global variable in this
# module.
_do_preparse = True


def preparser(on=True):
    """
    Turn on or off the Sage preparser.

    - ``on`` -- boolean; whether to turn on preparsing

    EXAMPLES::

        sage: 2/3
        2/3
        sage: preparser(False)
        sage: 2/3  # not tested since doctests are always preparsed
        0
        sage: preparser(True)
        sage: 2^3
        8
    """
    global _do_preparse
    _do_preparse = on is True


##############################
# Sage[Terminal]InteractiveShell
##############################
class SageShellOverride:
    """
    Mixin to override methods in IPython's [Terminal]InteractiveShell
    classes.
    """

    def show_usage(self):
        """
        Print the basic Sage usage.

        This method ends up being called when you enter ``?`` and
        nothing else on the command line.

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: shell.run_cell('?')
            Welcome to Sage ...
            sage: shell.quit()
        """
        from sage.misc.sagedoc import help
        help()

    def system_raw(self, cmd):
        """
        Run a system command.

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: shell.system_raw('false')
            sage: shell.user_ns['_exit_code'] > 0
            True
            sage: shell.system_raw('true')
            sage: shell.user_ns['_exit_code']
            0
            sage: shell.system_raw('R --version')   # optional - r
            R version ...
            sage: shell.user_ns['_exit_code']       # optional - r
            0
            sage: shell.quit()
        """
        return super().system_raw(cmd)


class SageNotebookInteractiveShell(SageShellOverride, InteractiveShell):
    """
    IPython Shell for the Sage IPython Notebook.

    The doctests are not tested since they would change the current
    rich output backend away from the doctest rich output backend.

    EXAMPLES::

        sage: from sage.repl.interpreter import SageNotebookInteractiveShell
        sage: SageNotebookInteractiveShell()   # not tested
        <sage.repl.interpreter.SageNotebookInteractiveShell object at 0x...>
    """

    def init_display_formatter(self):
        """
        Switch to the Sage IPython notebook rich output backend.

        EXAMPLES::

            sage: from sage.repl.interpreter import SageNotebookInteractiveShell
            sage: SageNotebookInteractiveShell().init_display_formatter()   # not tested
        """
        from sage.repl.rich_output.backend_ipython import BackendIPythonNotebook
        backend = BackendIPythonNotebook()
        backend.get_display_manager().switch_backend(backend, shell=self)


class SageTerminalInteractiveShell(SageShellOverride, TerminalInteractiveShell):
    """
    IPython Shell for the Sage IPython Commandline Interface.

    The doctests are not tested since they would change the current
    rich output backend away from the doctest rich output backend.

    EXAMPLES::

        sage: from sage.repl.interpreter import SageTerminalInteractiveShell
        sage: SageTerminalInteractiveShell()   # not tested
        <sage.repl.interpreter.SageNotebookInteractiveShell object at 0x...>
    """

    def init_display_formatter(self):
        """
        Switch to the Sage IPython commandline rich output backend.

        EXAMPLES::

            sage: from sage.repl.interpreter import SageTerminalInteractiveShell
            sage: SageTerminalInteractiveShell().init_display_formatter()   # not tested
        """
        from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
        backend = BackendIPythonCommandline()
        backend.get_display_manager().switch_backend(backend, shell=self)

    def prompt_for_code(self):
        # save sigint handlers (python and os level)
        # https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1576
        # https://github.com/sagemath/sage/issues/33428
        # https://github.com/sagemath/sage/pull/35251
        import signal
        sigint = signal.getsignal(signal.SIGINT)
        sigint_os = pythonapi.PyOS_getsig(signal.SIGINT)
        text = TerminalInteractiveShell.prompt_for_code(self)
        signal.signal(signal.SIGINT, sigint)
        pythonapi.PyOS_setsig(signal.SIGINT, sigint_os)
        return text


class SageTestShell(SageShellOverride, TerminalInteractiveShell):
    """
    Test Shell.

    Care must be taken in these doctests to quit the test shell in
    order to switch back the rich output display backend to the
    doctest backend.

    EXAMPLES::

        sage: from sage.repl.interpreter import get_test_shell
        sage: shell = get_test_shell();  shell
        <sage.repl.interpreter.SageTestShell object at 0x...>
        sage: shell.quit()
    """

    def init_display_formatter(self):
        """
        Switch to the Sage IPython commandline rich output backend.

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell();  shell
            <sage.repl.interpreter.SageTestShell object at 0x...>
            sage: shell.quit()
            sage: shell.init_display_formatter()
            sage: shell.quit()
        """
        from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
        self._ipython_backend = backend = BackendIPythonCommandline()
        self._display_manager = backend.get_display_manager()
        self._doctest_backend = self._display_manager.switch_backend(backend, shell=self)

    def quit(self):
        """
        Quit the test shell.

        To make the test shell as realistic as possible, we switch to
        the
        :class:`~sage.repl.rich_output.backend_ipython.BackendIPythonCommandline`
        display backend. This method restores the previous display
        backend, which is the
        :class:`~sage.repl.rich_output.backend_doctest.BackendDoctest`
        during doctests.

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: from sage.repl.rich_output import get_display_manager
            sage: get_display_manager()
            The Sage display manager using the doctest backend

            sage: shell = get_test_shell()
            sage: get_display_manager()
            The Sage display manager using the IPython command line backend

            sage: shell.quit()
            sage: get_display_manager()
            The Sage display manager using the doctest backend
        """
        self._display_manager.switch_backend(self._doctest_backend)

    def _restart(self):
        """
        Restart the test shell (after :meth:`quit`).

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: shell.quit()
            sage: shell._restart()
            sage: shell.quit()
            sage: from sage.repl.rich_output import get_display_manager
            sage: get_display_manager()
            The Sage display manager using the doctest backend
        """
        self._display_manager.switch_backend(self._ipython_backend, shell=self)

    def run_cell(self, *args, **kwds):
        """
        Run IPython cell.

        Starting with IPython-3.0, this returns an success/failure
        information. Since it is more convenient for doctests, we
        ignore it.

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: rc = shell.run_cell('2^50')
            1125899906842624
            sage: rc is None
            True
            sage: shell.quit()
        """
        super().run_cell(*args, **kwds)


###################################################################
# Transformers used in the SageInputSplitter
###################################################################


def SagePreparseTransformer(lines):
    r"""
    EXAMPLES::

        sage: from sage.repl.interpreter import SagePreparseTransformer
        sage: spt = SagePreparseTransformer
        sage: spt(['1+1r+2.3^2.3r\n'])
        ["Integer(1)+1+RealNumber('2.3')**2.3\n"]
        sage: preparser(False)
        sage: spt(['2.3^2\n'])
        ['2.3^2\n']

    TESTS:

    Check that syntax errors in the preparser do not crash IPython,
    see :issue:`14961`. ::

        sage: preparser(True)
        sage: bad_syntax = "R.<t> = QQ{]"
        sage: preparse(bad_syntax)
        Traceback (most recent call last):
        ...
        SyntaxError: mismatched ']'
        sage: from sage.repl.interpreter import get_test_shell
        sage: shell = get_test_shell()
        sage: shell.run_cell(bad_syntax)
          File...<string>...
        SyntaxError: mismatched ']'
        <BLANKLINE>
        sage: shell.quit()

    Make sure the quote state is carried over across subsequent lines in order
    to avoid interfering with multi-line strings, see :issue:`30417`. ::

        sage: SagePreparseTransformer(["'''\n", 'abc-1-2\n', "'''\n"])
        ["'''\n", 'abc-1-2\n', "'''\n"]
        sage: # instead of ["'''\n", 'abc-Integer(1)-Integer(2)\n', "'''\n"]

    .. NOTE::

        IPython may call this function more than once for the same input lines.
        So when debugging the preparser, print outs may be duplicated. If using
        IPython >= 7.17, try:
        ``sage.repl.interpreter.SagePreparseTransformer.has_side_effects = True``
    """
    if _do_preparse:
        # IPython ensures the input lines end with a newline, and it expects
        # the same of the output lines.
        lines = preparse(''.join(lines)).splitlines(keepends=True)
    return lines


SagePromptTransformer = PromptStripper(prompt_re=re.compile(r'^(\s*(:?sage: |\.\.\.\.: ))+'))


###################
# Interface shell #
###################

class logstr(str):
    """
    For use by :meth`~InterfaceShellTransformer.transform`.
    This provides a ``_latex_`` method which is just the string
    wrapped in a ``\\verb`` environment.
    """
    def __repr__(self):
        """
        EXAMPLES::
            sage: from sage.repl.interpreter import logstr
            sage: logstr("ABC")
            ABC
        """
        return self

    def _latex_(self):
        r"""
        EXAMPLES::
            sage: from sage.repl.interpreter import logstr
            sage: logstr("A")
            A
            sage: latex(logstr("A"))
            \verb#A#
            sage: latex(logstr("A#B"))
            \verb@A#B@
        """
        # return "\\begin{verbatim}%s\\end{verbatim}"%self
        if '#' not in self:
            delim = '#'
        elif '@' not in self:
            delim = '@'
        elif '~' not in self:
            delim = '~'
        return r"""\verb%s%s%s""" % (delim, self.replace('\n\n', '\n').replace('\n', '; '), delim)


class InterfaceShellTransformer(PrefilterTransformer):
    priority = 50

    def __init__(self, *args, **kwds):
        """
        Initialize this class.  All of the arguments get passed to
        :meth:`PrefilterTransformer.__init__`.

        .. attribute:: temporary_objects

           a list of hold onto interface objects and keep them from being
           garbage collected

        .. SEEALSO:: :func:`interface_shell_embed`

        TESTS::

            sage: # needs sage.symbolic
            sage: from sage.repl.interpreter import interface_shell_embed
            sage: shell = interface_shell_embed(maxima)
            sage: ift = shell.prefilter_manager.transformers[0]
            sage: ift.temporary_objects
            set()
            sage: ift._sage_import_re.findall('sage(a) + maxima(b)')
            ['sage(', 'maxima(']
        """
        super().__init__(*args, **kwds)
        self.temporary_objects = set()
        self._sage_import_re = re.compile(r'(?:sage|%s)\('
                                          % self.shell.interface.name())

    def preparse_imports_from_sage(self, line):
        """
        Finds occurrences of strings such as ``sage(object)`` in
        *line*, converts ``object`` to :attr:`shell.interface`,
        and replaces those strings with their identifier in the new
        system.  This also works with strings such as
        ``maxima(object)`` if :attr:`shell.interface` is
        ``maxima``.

        - ``line`` -- string; the line to transform

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: from sage.repl.interpreter import interface_shell_embed, InterfaceShellTransformer
            sage: shell = interface_shell_embed(maxima)
            sage: ift = InterfaceShellTransformer(shell=shell, config=shell.config,
            ....:     prefilter_manager=shell.prefilter_manager)
            sage: ift.shell.ex('a = 3')
            sage: ift.preparse_imports_from_sage('2 + sage(a)')
            '2 + sage0 '
            sage: maxima.eval('sage0')
            '3'
            sage: ift.preparse_imports_from_sage('2 + maxima(a)') # maxima calls set_seed on startup which is why 'sage0' will becomes 'sage4' and not just 'sage1'
            '2 +  sage4 '
            sage: ift.preparse_imports_from_sage('2 + gap(a)')
            '2 + gap(a)'

        Since :issue:`28439`, this also works with more complicated expressions
        containing nested parentheses::

            sage: # needs sage.libs.gap sage.symbolic
            sage: shell = interface_shell_embed(gap)
            sage: shell.user_ns = locals()
            sage: ift = InterfaceShellTransformer(shell=shell, config=shell.config,
            ....:     prefilter_manager=shell.prefilter_manager)
            sage: line = '2 + sage((1+2)*gap(-(5-3)^2).sage()) - gap(1+(2-1))'
            sage: line = ift.preparse_imports_from_sage(line)
            sage: gap.eval(line)
            '-12'
        """
        new_line = []
        pos = 0
        while True:
            m = self._sage_import_re.search(line, pos)
            if not m:
                new_line.append(line[pos:])
                break
            expr_start, expr_end = containing_block(line, m.end() - 1,
                                                    delimiters=['()'])
            expr = preparse(line[expr_start + 1:expr_end - 1])
            result = self.shell.interface(eval(expr, self.shell.user_ns))
            self.temporary_objects.add(result)
            new_line += [line[pos:m.start()], result.name()]
            pos = expr_end
        return ' '.join(new_line)

    def transform(self, line, continue_prompt):
        r'''
        Evaluates *line* in :attr:`shell.interface` and returns a
        string representing the result of that evaluation.

        - ``line`` -- string; the line to be transformed *and evaluated*
        - ``continue_prompt`` -- boolean; whether this line is a continuation in a
          sequence of multiline input

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: from sage.repl.interpreter import interface_shell_embed, InterfaceShellTransformer
            sage: shell = interface_shell_embed(maxima)
            sage: ift = InterfaceShellTransformer(shell=shell, config=shell.config,
            ....:     prefilter_manager=shell.prefilter_manager)
            sage: ift.transform('2+2', False)   # note: output contains triple quotation marks
            'sage.repl.interpreter.logstr(r"""4""")'
            sage: ift.shell.ex('a = 4')
            sage: ift.transform(r'sage(a)+4', False)
            'sage.repl.interpreter.logstr(r"""8""")'
            sage: ift.temporary_objects
            set()
            sage: shell = interface_shell_embed(gap)                                    # needs sage.libs.gap
            sage: ift = InterfaceShellTransformer(shell=shell, config=shell.config,     # needs sage.libs.gap
            ....:     prefilter_manager=shell.prefilter_manager)
            sage: ift.transform('2+2', False)
            'sage.repl.interpreter.logstr(r"""4""")'

        TESTS:

        Check that whitespace is not stripped and that special characters are
        escaped (:issue:`28439`)::

            sage: shell = interface_shell_embed(gap)                                    # needs sage.libs.gap sage.symbolic
            sage: ift = InterfaceShellTransformer(shell=shell, config=shell.config,     # needs sage.libs.gap sage.symbolic
            ....:     prefilter_manager=shell.prefilter_manager)
            sage: ift.transform(r'Print("  -\n\\\\-  ");', False)                       # needs sage.symbolic
            'sage.repl.interpreter.logstr(r"""  -\n\\\\-""")'

            sage: # optional - macaulay2
            sage: shell = interface_shell_embed(macaulay2)
            sage: ift = InterfaceShellTransformer(shell=shell, config=shell.config,
            ....:     prefilter_manager=shell.prefilter_manager)
            sage: ift.transform('net(ZZ^2)', False)
            'sage.repl.interpreter.logstr(r"""  2\nZZ""")'
        '''
        line = self.preparse_imports_from_sage(line)

        try:
            t = self.shell.interface.eval(line)
        finally:
            # Once we've evaluated the lines, we can clear the
            # temporary objects
            self.temporary_objects = set()
        # We do not strip whitespace from t here as the individual interface is
        # responsible for that
        return 'sage.repl.interpreter.logstr(r"""%s""")' % t


def interface_shell_embed(interface):
    """
    Return an IPython shell which uses a Sage interface on the
    backend to perform the evaluations.  It uses
    :class:`InterfaceShellTransformer` to transform the input into the
    appropriate ``interface.eval(...)`` input.

    INPUT:

    - ``interface`` -- a Sage ``PExpect`` interface instance

    EXAMPLES::

        sage: from sage.repl.interpreter import interface_shell_embed
        sage: shell = interface_shell_embed(gap)                                        # needs sage.libs.gap
        sage: shell.run_cell('List( [1..10], IsPrime )')                                # needs sage.libs.gap
        [ false, true, true, false, true, false, true, false, false, false ]
        <ExecutionResult object at ..., execution_count=None error_before_exec=None error_in_exec=None ...result=[ false, true, true, false, true, false, true, false, false, false ]>
    """
    cfg = sage_ipython_config.copy()
    ipshell = InteractiveShellEmbed(config=cfg,
                                    banner1='\n  --> Switching to %s <--\n\n' % interface,
                                    exit_msg='\n  --> Exiting back to Sage <--\n')
    ipshell.interface = interface
    ipshell.prompts = InterfacePrompts(interface.name())

    while ipshell.prefilter_manager.transformers:
        ipshell.prefilter_manager.transformers.pop()
    while ipshell.prefilter_manager.checkers:
        ipshell.prefilter_manager.checkers.pop()
    ipshell.ex('import sage.misc.all')

    InterfaceShellTransformer(shell=ipshell,
                              prefilter_manager=ipshell.prefilter_manager,
                              config=cfg)
    return ipshell


def get_test_shell():
    """
    Return a IPython shell that can be used in testing the functions
    in this module.

    OUTPUT: an IPython shell

    EXAMPLES::

        sage: from sage.repl.interpreter import get_test_shell
        sage: shell = get_test_shell(); shell
        <sage.repl.interpreter.SageTestShell object at 0x...>
        sage: shell.parent.shell_class
        <class 'sage.repl.interpreter.SageTestShell'>
        sage: shell.parent.test_shell
        True
        sage: shell.quit()

    TESTS:

    Check that :issue:`14070` has been resolved::

        sage: from sage.tests.cmdline import test_executable
        sage: cmd = 'from sage.repl.interpreter import get_test_shell; shell = get_test_shell()'
        sage: (out, err, ret) = test_executable(["sage", "-c", cmd])
        sage: out + err
        ''
    """
    config = sage_ipython_config.default()
    config.TerminalIPythonApp.test_shell = True
    config.TerminalIPythonApp.shell_class = SageTestShell
    app = SageTerminalApp.instance(config=config)
    if app.shell is None:
        app.initialize(argv=[])
    else:
        try:
            app.shell._restart()
        except AttributeError:
            pass
    # No quit noise
    app.shell.verbose_quit = False
    return app.shell


#######################
# IPython TerminalApp #
#######################

class SageCrashHandler(IPAppCrashHandler):
    def __init__(self, app):
        """
        A custom :class:`CrashHandler` which gives the user
        instructions on how to post the problem to sage-support.

        EXAMPLES::

            sage: from sage.repl.interpreter import SageTerminalApp, SageCrashHandler
            sage: app = SageTerminalApp.instance()
            sage: sch = SageCrashHandler(app); sch
            <sage.repl.interpreter.SageCrashHandler object at 0x...>
            sage: sorted(sch.info.items())
            [('app_name', 'Sage'),
             ('bug_tracker', 'https://github.com/sagemath/sage/issues'),
             ('contact_email', 'sage-support@googlegroups.com'),
             ('contact_name', 'sage-support'),
             ('crash_report_fname', 'Crash_report_Sage.txt')]
        """
        contact_name = 'sage-support'
        contact_email = 'sage-support@googlegroups.com'
        bug_tracker = 'https://github.com/sagemath/sage/issues'
        CrashHandler.__init__(self,
                              app, contact_name, contact_email,
                              bug_tracker, show_crash_traceback=True)
        self.crash_report_fname = 'Sage_crash_report.txt'


class SageTerminalApp(TerminalIPythonApp):
    name = 'Sage'
    crash_handler_class = SageCrashHandler

    test_shell = Bool(False, help='Whether the shell is a test shell')
    test_shell.tag(config=True)
    shell_class = Type(InteractiveShell, help='Type of the shell')
    shell_class.tag(config=True)

    def load_config_file(self, *args, **kwds):
        r"""
        Merges a config file with the default sage config.

        .. NOTE::

            This code is based on :meth:`Application.update_config`.

        TESTS:

        Test that :issue:`15972` has been fixed::

            sage: import tempfile
            sage: from sage.repl.interpreter import SageTerminalApp
            sage: from IPython.paths import get_ipython_dir
            sage: with tempfile.TemporaryDirectory() as d:
            ....:     IPYTHONDIR = get_ipython_dir()
            ....:     os.environ['IPYTHONDIR'] = d
            ....:     SageTerminalApp().load_config_file()
            ....:     os.environ['IPYTHONDIR'] = IPYTHONDIR
        """
        super().load_config_file(*args, **kwds)
        newconfig = sage_ipython_config.default()
        # merge in the config loaded from file
        newconfig.merge(self.config)
        self.config = newconfig

    def init_shell(self):
        r"""
        Initialize the :class:`SageInteractiveShell` instance.

        .. NOTE::

            This code is based on
            :meth:`TerminalIPythonApp.init_shell`.

        EXAMPLES::

            sage: from sage.repl.interpreter import SageTerminalApp
            sage: app = SageTerminalApp.instance()
            sage: app.shell
            <sage.repl.interpreter.SageTestShell object at 0x...>
        """
        # Shell initialization
        self.shell = self.shell_class.instance(
            parent=self,
            config=self.config,
            profile_dir=self.profile_dir,
            ipython_dir=self.ipython_dir)
        self.shell.configurables.append(self)
        self.shell.has_sage_extensions = SAGE_EXTENSION in self.extensions

        # Load the %lprun extension if available
        try:
            import line_profiler
            assert line_profiler  # silence pyflakes
        except ImportError:
            pass
        else:
            self.extensions.append('line_profiler')

        if self.shell.has_sage_extensions:
            self.extensions.remove(SAGE_EXTENSION)

            # load sage extension here to get a crash if
            # something is wrong with the sage library
            self.shell.extension_manager.load_extension(SAGE_EXTENSION)



# sage_setup: distribution = sagemath-repl
"""
Classes for sources of doctests

This module defines various classes for sources from which doctests
originate, such as files, functions or database entries.

AUTHORS:

- David Roe (2012-03-27) -- initial version, based on Robert Bradshaw's code.
"""

# ****************************************************************************
#       Copyright (C) 2012-2013 David Roe <roed.math@gmail.com>
#                     2012      Robert Bradshaw <robertwb@gmail.com>
#                     2012      William Stein <wstein@gmail.com>
#                     2013      R. Andrew Ohana
#                     2013-2017 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#                     2013-2019 John H. Palmieri
#                     2014      Volker Braun
#                     2014-2022 Frédéric Chapoton
#                     2017      Erik M. Bray
#                     2021      Sébastien Labbé
#                     2021-2023 Matthias Koeppe
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os
import sys
import re
import random
import doctest
from sage.cpython.string import bytes_to_str
from sage.repl.load import load
from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.package_dir import is_package_or_sage_namespace_package_dir
from .parsing import SageDocTestParser
from .util import NestedName
from sage.structure.dynamic_class import dynamic_class
from sage.env import SAGE_SRC, SAGE_LIB

# Python file parsing
triple_quotes = re.compile(r"\s*[rRuU]*((''')|(\"\"\"))")
name_regex = re.compile(r".*\s(\w+)([(].*)?:")

# LaTeX file parsing
begin_verb = re.compile(r"\s*\\begin{verbatim}")
end_verb = re.compile(r"\s*\\end{verbatim}\s*(%link)?")
begin_lstli = re.compile(r"\s*\\begin{lstlisting}")
end_lstli = re.compile(r"\s*\\end{lstlisting}\s*(%link)?")
skip = re.compile(r".*%skip.*")

# ReST file parsing
link_all = re.compile(r"^\s*\.\.\s+linkall\s*$")
double_colon = re.compile(r"^(\s*).*::\s*$")
code_block = re.compile(r"^(\s*)[.][.]\s*code-block\s*::.*$")

whitespace = re.compile(r"\s*")
bitness_marker = re.compile('#.*(32|64)-bit')
bitness_value = '64' if sys.maxsize > (1 << 32) else '32'

# For neutralizing doctests
find_prompt = re.compile(r"^(\s*)(>>>|sage:)(.*)")

# For testing that enough doctests are created
sagestart = re.compile(r"^(\s*(>>> |sage: ))\s*[^#\s]")
untested = re.compile("(not implemented|not tested)")

# For parsing a PEP 0263 encoding declaration
pep_0263 = re.compile(br'^[ \t\v]*#.*?coding[:=]\s*([-\w.]+)')

# Source line number in warning output
doctest_line_number = re.compile(r"^\s*doctest:[0-9]")


def get_basename(path):
    """
    This function returns the basename of the given path, e.g.
    ``sage.doctest.sources`` or ``doc.ru.tutorial.tour_advanced``.

    EXAMPLES::

        sage: from sage.doctest.sources import get_basename
        sage: import os
        sage: get_basename(sage.doctest.sources.__file__)
        'sage.doctest.sources'

    ::

        sage: # optional - !meson_editable
        sage: get_basename(os.path.join(sage.structure.__path__[0], 'element.pxd'))
        'sage.structure.element.pxd'

    TESTS::

        sage: # optional - meson_editable
        sage: get_basename(os.path.join(os.path.dirname(sage.structure.__file__), 'element.pxd'))
        'sage.structure.element.pxd'
    """
    if path is None:
        return None
    if not os.path.exists(path):
        return path
    path = os.path.abspath(path)
    root = os.path.dirname(path)
    # If the file is in the sage library, we can use our knowledge of
    # the directory structure
    dev = SAGE_SRC
    sp = SAGE_LIB
    if path.startswith(dev):
        # there will be a branch name
        i = path.find(os.path.sep, len(dev))
        if i == -1:
            # this source is the whole library....
            return path
        root = path[:i]
    elif path.startswith(sp):
        root = path[:len(sp)]
    else:
        # If this file is in some python package we can see how deep
        # it goes.
        while is_package_or_sage_namespace_package_dir(root):
            root = os.path.dirname(root)
    fully_qualified_path, ext = os.path.splitext(path[len(root) + 1:])
    if os.path.split(path)[1] == '__init__.py':
        fully_qualified_path = fully_qualified_path[:-9]
    basename = fully_qualified_path.replace(os.path.sep, '.')
    if ext in ['.pxd', '.pxi']:
        # disambiguate from .pyx with the same basename
        basename += ext
    return basename


class DocTestSource:
    """
    This class provides a common base class for different sources of doctests.

    INPUT:

    - ``options`` -- a :class:`sage.doctest.control.DocTestDefaults`
      instance or equivalent
    """
    def __init__(self, options):
        """
        Initialization.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: TestSuite(FDS).run()
        """
        self.options = options

    def __eq__(self, other):
        """
        Comparison is just by comparison of attributes.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.doctest.sources.__file__
            sage: DD = DocTestDefaults()
            sage: FDS = FileDocTestSource(filename, DD)
            sage: FDS2 = FileDocTestSource(filename, DD)
            sage: FDS == FDS2
            True
        """
        if type(self) is not type(other):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Test for non-equality.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.doctest.sources.__file__
            sage: DD = DocTestDefaults()
            sage: FDS = FileDocTestSource(filename, DD)
            sage: FDS2 = FileDocTestSource(filename, DD)
            sage: FDS != FDS2
            False
        """
        return not (self == other)

    def _process_doc(self, doctests, doc, namespace, start):
        """
        Appends doctests defined in ``doc`` to the list ``doctests``.

        This function is called when a docstring block is completed
        (either by ending a triple quoted string in a Python file,
        unindenting from a comment block in a ReST file, or ending a
        verbatim or lstlisting environment in a LaTeX file).

        INPUT:

        - ``doctests`` -- a running list of doctests to which the new
          test(s) will be appended

        - ``doc`` -- list of lines of a docstring, each including
          the trailing newline

        - ``namespace`` -- dictionary or
          :class:`sage.doctest.util.RecordingDict`, used in the
          creation of new :class:`doctest.DocTest` s

        - ``start`` -- integer giving the line number of the start
          of this docstring in the larger file

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: from sage.doctest.parsing import SageDocTestParser
            sage: filename = sage.doctest.util.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: doctests, _ = FDS.create_doctests({})
            sage: manual_doctests = []
            sage: for dt in doctests:
            ....:     FDS.qualified_name = dt.name
            ....:     FDS._process_doc(manual_doctests, dt.docstring, {}, dt.lineno-1)
            sage: doctests == manual_doctests
            True
        """
        docstring = "".join(doc)
        new_doctests = self.parse_docstring(docstring, namespace, start)
        sig_on_count_doc_doctest = "sig_on_count() # check sig_on/off pairings (virtual doctest)\n"
        for dt in new_doctests:
            if len(dt.examples) > 0 and not (hasattr(dt.examples[-1], 'sage_source')
                                             and dt.examples[-1].sage_source == sig_on_count_doc_doctest):
                # Line number refers to the end of the docstring
                sigon = doctest.Example(sig_on_count_doc_doctest, "0\n", lineno=docstring.count("\n"))
                sigon.sage_source = sig_on_count_doc_doctest
                sigon.optional_tags = frozenset(self.file_optional_tags)
                sigon.probed_tags = frozenset()
                dt.examples.append(sigon)
            doctests.append(dt)

    @lazy_attribute
    def file_optional_tags(self):
        r"""
        Return the set of tags that should apply to all doctests in this source.

        This default implementation just returns the empty set.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import StringDocTestSource, PythonSource
            sage: from sage.structure.dynamic_class import dynamic_class
            sage: s = "'''\n    sage: 2 + 2\n    4\n'''"
            sage: PythonStringSource = dynamic_class('PythonStringSource', (StringDocTestSource, PythonSource))
            sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
            sage: PSS.file_optional_tags
            set()
        """
        return set()

    def _create_doctests(self, namespace, tab_okay=None):
        """
        Create a list of doctests defined in this source.

        This function collects functionality common to file and string
        sources, and is called by
        :meth:`FileDocTestSource.create_doctests`.

        INPUT:

        - ``namespace`` -- dictionary or
          :class:`sage.doctest.util.RecordingDict`, used in the
          creation of new :class:`doctest.DocTest` s.

        - ``tab_okay`` -- whether tabs are allowed in this source

        OUTPUT:

        - ``doctests`` -- list of doctests defined by this source

        - ``extras`` -- dictionary with ``extras['tab']`` either
          ``False`` or a list of linenumbers on which tabs appear

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: from sage.doctest.util import NestedName
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS.qualified_name = NestedName('sage.doctest.sources')
            sage: doctests, extras = FDS._create_doctests({})
            sage: len(doctests)
            43
            sage: extras['tab']
            False
            sage: extras['line_number']
            False
        """
        if tab_okay is None:
            tab_okay = isinstance(self, TexSource)
        self._init()
        self.line_shift = 0
        self.parser = SageDocTestParser(self.options.optional,
                                        self.options.long,
                                        probed_tags=self.options.probe,
                                        file_optional_tags=self.file_optional_tags)
        self.linking = False
        doctests = []
        in_docstring = False
        unparsed_doc = False
        doc = []
        start = None
        tab_locations = []
        contains_line_number = False
        for lineno, line in self:
            if doctest_line_number.search(line) is not None:
                contains_line_number = True
            if "\t" in line:
                tab_locations.append(str(lineno+1))
            if "SAGE_DOCTEST_ALLOW_TABS" in line:
                tab_okay = True
            just_finished = False
            if in_docstring:
                if self.ending_docstring(line):
                    in_docstring = False
                    just_finished = True
                    self._process_doc(doctests, doc, namespace, start)
                    unparsed_doc = False
                else:
                    bitness = bitness_marker.search(line)
                    if bitness:
                        if bitness.groups()[0] != bitness_value:
                            self.line_shift += 1
                            continue
                        else:
                            line = line[:bitness.start()] + "\n"
                    if self.line_shift and (m := sagestart.match(line)):
                        # We insert empty doctest lines to make up for the removed lines
                        indent_and_prompt = m.group(1)
                        doc.extend([indent_and_prompt + "# inserted to compensate for removed conditional doctest output\n"]
                                   * self.line_shift)
                        self.line_shift = 0
                    doc.append(line)
                    unparsed_doc = True
            if not in_docstring and (not just_finished or self.start_finish_can_overlap):
                # to get line numbers in linked docstrings correct we
                # append a blank line to the doc list.
                doc.append("\n")
                if not line.strip():
                    continue
                if self.starting_docstring(line):
                    in_docstring = True
                    if self.linking:
                        # If there's already a doctest, we overwrite it.
                        if len(doctests) > 0:
                            doctests.pop()
                        if start is None:
                            start = lineno
                            doc = []
                    else:
                        self.line_shift = 0
                        start = lineno
                        doc = []
        # In ReST files we can end the file without decreasing the indentation level.
        if unparsed_doc:
            self._process_doc(doctests, doc, namespace, start)

        extras = {"tab": not tab_okay and tab_locations,
                  "line_number": contains_line_number,
                  "optionals": self.parser.optionals}
        if self.options.randorder is not None and self.options.randorder is not False:
            # we want to randomize even when self.randorder = 0
            random.seed(self.options.randorder)
            randomized = []
            while doctests:
                i = random.randint(0, len(doctests) - 1)
                randomized.append(doctests.pop(i))
            return randomized, extras
        else:
            return doctests, extras


class StringDocTestSource(DocTestSource):
    r"""
    This class creates doctests from a string.

    INPUT:

    - ``basename`` -- string such as 'sage.doctests.sources', going
      into the names of created doctests and examples

    - ``source`` -- string, giving the source code to be parsed for
      doctests

    - ``options`` -- a :class:`sage.doctest.control.DocTestDefaults`
      or equivalent

    - ``printpath`` -- string, to be used in place of a filename
      when doctest failures are displayed

    - ``lineno_shift`` -- integer (default: 0) by which to shift
      the line numbers of all doctests defined in this string

    EXAMPLES::

        sage: from sage.doctest.control import DocTestDefaults
        sage: from sage.doctest.sources import StringDocTestSource, PythonSource
        sage: from sage.structure.dynamic_class import dynamic_class
        sage: s = "'''\n    sage: 2 + 2\n    4\n'''"
        sage: PythonStringSource = dynamic_class('PythonStringSource',(StringDocTestSource, PythonSource))
        sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
        sage: dt, extras = PSS.create_doctests({})
        sage: len(dt)
        1
        sage: extras['tab']
        []
        sage: extras['line_number']
        False

        sage: s = "'''\n\tsage: 2 + 2\n\t4\n'''"
        sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
        sage: dt, extras = PSS.create_doctests({})
        sage: extras['tab']
        ['2', '3']

        sage: s = "'''\n    sage: import warnings; warnings.warn('foo')\n    doctest:1: UserWarning: foo \n'''"
        sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
        sage: dt, extras = PSS.create_doctests({})
        sage: extras['line_number']
        True
    """
    def __init__(self, basename, source, options, printpath, lineno_shift=0):
        r"""
        Initialization.

        TESTS::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import StringDocTestSource, PythonSource
            sage: from sage.structure.dynamic_class import dynamic_class
            sage: s = "'''\n    sage: 2 + 2\n    4\n'''"
            sage: PythonStringSource = dynamic_class('PythonStringSource',(StringDocTestSource, PythonSource))
            sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
            sage: TestSuite(PSS).run()
        """
        self.qualified_name = NestedName(basename)
        self.printpath = printpath
        self.source = source
        self.lineno_shift = lineno_shift
        DocTestSource.__init__(self, options)

    def __iter__(self):
        r"""
        Iterating over this source yields pairs ``(lineno, line)``.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import StringDocTestSource, PythonSource
            sage: from sage.structure.dynamic_class import dynamic_class
            sage: s = "'''\n    sage: 2 + 2\n    4\n'''"
            sage: PythonStringSource = dynamic_class('PythonStringSource',(StringDocTestSource, PythonSource))
            sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
            sage: for n, line in PSS:
            ....:     print("{} {}".format(n, line))
            0 '''
            1     sage: 2 + 2
            2     4
            3 '''
        """
        for lineno, line in enumerate(self.source.split('\n')):
            yield lineno + self.lineno_shift, line + '\n'

    def create_doctests(self, namespace):
        r"""
        Create doctests from this string.

        INPUT:

        - ``namespace`` -- dictionary or :class:`sage.doctest.util.RecordingDict`

        OUTPUT:

        - ``doctests`` -- list of doctests defined by this string

        - ``tab_locations`` -- either ``False`` or a list of linenumbers
          on which tabs appear

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import StringDocTestSource, PythonSource
            sage: from sage.structure.dynamic_class import dynamic_class
            sage: s = "'''\n    sage: 2 + 2\n    4\n'''"
            sage: PythonStringSource = dynamic_class('PythonStringSource',(StringDocTestSource, PythonSource))
            sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
            sage: dt, tabs = PSS.create_doctests({})
            sage: for t in dt:
            ....:     print("{} {}".format(t.name, t.examples[0].sage_source))
            <runtime> 2 + 2
        """
        return self._create_doctests(namespace)


class FileDocTestSource(DocTestSource):
    """
    This class creates doctests from a file.

    INPUT:

    - ``path`` -- string; the filename

    - ``options`` -- a :class:`sage.doctest.control.DocTestDefaults`
      instance or equivalent

    EXAMPLES::

        sage: from sage.doctest.control import DocTestDefaults
        sage: from sage.doctest.sources import FileDocTestSource
        sage: filename = sage.doctest.sources.__file__
        sage: FDS = FileDocTestSource(filename, DocTestDefaults())
        sage: FDS.basename
        'sage.doctest.sources'

    TESTS::

        sage: TestSuite(FDS).run()

    ::

        sage: from sage.doctest.control import DocTestDefaults
        sage: from sage.doctest.sources import FileDocTestSource
        sage: filename = tmp_filename(ext='.txtt')
        sage: FDS = FileDocTestSource(filename, DocTestDefaults())
        Traceback (most recent call last):
        ...
        ValueError: unknown extension for the file to test (=...txtt),
        valid extensions are: .py, .pyx, .pxd, .pxi, .sage, .spyx, .tex, .rst, .rst.txt
    """
    def __init__(self, path, options):
        """
        Initialization.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults(randorder=0))
            sage: FDS.options.randorder
            0
        """
        self.path = path
        DocTestSource.__init__(self, options)
        if path.endswith('.rst.txt'):
            ext = '.rst.txt'
        else:
            base, ext = os.path.splitext(path)
        valid_code_ext = ('.py', '.pyx', '.pxd', '.pxi', '.sage', '.spyx')
        if ext in valid_code_ext:
            self.__class__ = dynamic_class('PythonFileSource', (FileDocTestSource, PythonSource))
            self.encoding = "utf-8"
        elif ext == '.tex':
            self.__class__ = dynamic_class('TexFileSource', (FileDocTestSource, TexSource))
            self.encoding = "utf-8"
        elif ext == '.rst' or ext == '.rst.txt':
            self.__class__ = dynamic_class('RestFileSource', (FileDocTestSource, RestSource))
            self.encoding = "utf-8"
        else:
            valid_ext = ", ".join(valid_code_ext + ('.tex', '.rst', '.rst.txt'))
            raise ValueError("unknown extension for the file to test (={}),"
                    " valid extensions are: {}".format(path, valid_ext))

    def __iter__(self):
        r"""
        Iterating over this source yields pairs ``(lineno, line)``.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = tmp_filename(ext='.py')
            sage: s = "'''\n    sage: 2 + 2\n    4\n'''"
            sage: with open(filename, 'w') as f:
            ....:     _ = f.write(s)
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: for n, line in FDS:
            ....:     print("{} {}".format(n, line))
            0 '''
            1     sage: 2 + 2
            2     4
            3 '''

        The encoding is "utf-8" by default::

            sage: FDS.encoding
            'utf-8'

        We create a file with a Latin-1 encoding without declaring it::

            sage: s = b"'''\nRegardons le polyn\xF4me...\n'''\n"
            sage: with open(filename, 'wb') as f:
            ....:     _ = f.write(s)
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: L = list(FDS)
            Traceback (most recent call last):
            ...
            UnicodeDecodeError: 'utf...8' codec can...t decode byte 0xf4 in position 18: invalid continuation byte

        This works if we add a PEP 0263 encoding declaration::

            sage: s = b"#!/usr/bin/env python\n# -*- coding: latin-1 -*-\n" + s
            sage: with open(filename, 'wb') as f:
            ....:     _ = f.write(s)
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: L = list(FDS)
            sage: FDS.encoding
            'latin-1'
        """
        with open(self.path, 'rb') as source:
            for lineno, line in enumerate(source):
                if lineno < 2:
                    match = pep_0263.search(line)
                    if match:
                        self.encoding = bytes_to_str(match.group(1), 'ascii')
                yield lineno, line.decode(self.encoding)

    @lazy_attribute
    def printpath(self):
        """
        Whether the path is printed absolutely or relatively depends on an option.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: import os
            sage: filename = os.path.realpath(sage.doctest.sources.__file__)
            sage: root = os.path.join(os.path.dirname(filename), '..')
            sage: cwd = os.getcwd()
            sage: os.chdir(root)
            sage: FDS = FileDocTestSource(filename, DocTestDefaults(randorder=0,
            ....:                                                   abspath=False))
            sage: FDS.printpath
            'doctest/sources.py'
            sage: FDS = FileDocTestSource(filename, DocTestDefaults(randorder=0,
            ....:                                                   abspath=True))
            sage: FDS.printpath
            '.../sage/doctest/sources.py'
            sage: os.chdir(cwd)
        """
        if self.options.abspath:
            return os.path.abspath(self.path)
        else:
            relpath = os.path.relpath(self.path)
            if relpath.startswith(".." + os.path.sep):
                return self.path
            else:
                return relpath

    @lazy_attribute
    def basename(self):
        """
        The basename of this file source, e.g. ``sage.doctest.sources``.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = os.path.join(SAGE_SRC,'sage','rings','integer.pyx')
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS.basename
            'sage.rings.integer'
        """
        return get_basename(self.path)

    @lazy_attribute
    def in_lib(self):
        """
        Whether this file is to be treated as a module in a Python package.

        Such files aren't loaded before running tests.

        This uses :func:`~sage.misc.package_dir.is_package_or_sage_namespace_package_dir`
        but can be overridden via :class:`~sage.doctest.control.DocTestDefaults`.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: import os
            sage: filename = os.path.join(SAGE_SRC, 'sage', 'rings', 'integer.pyx')
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS.in_lib
            True
            sage: filename = os.path.join(SAGE_SRC, 'sage', 'doctest', 'tests', 'abort.rst')
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS.in_lib
            False

        You can override the default::

            sage: FDS = FileDocTestSource("hello_world.py", DocTestDefaults())
            sage: FDS.in_lib
            False
            sage: FDS = FileDocTestSource("hello_world.py", DocTestDefaults(force_lib=True))
            sage: FDS.in_lib
            True
        """
        return (self.options.force_lib
                or is_package_or_sage_namespace_package_dir(os.path.dirname(self.path)))

    @lazy_attribute
    def file_optional_tags(self):
        """
        Return the set of tags that should apply to all doctests in this source.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.repl.user_globals.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS.file_optional_tags
            {'sage.modules': None}
        """
        from .parsing import parse_file_optional_tags
        return parse_file_optional_tags(self)

    def create_doctests(self, namespace):
        r"""
        Return a list of doctests for this file.

        INPUT:

        - ``namespace`` -- dictionary or :class:`sage.doctest.util.RecordingDict`

        OUTPUT:

        - ``doctests`` -- list of doctests defined in this file

        - ``extras`` -- dictionary

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: doctests, extras = FDS.create_doctests(globals())
            sage: len(doctests)
            43
            sage: extras['tab']
            False

        We give a self referential example::

            sage: doctests[20].name
            'sage.doctest.sources.FileDocTestSource.create_doctests'
            sage: doctests[20].examples[8].source
            'doctests[Integer(20)].examples[Integer(8)].source\n'

        TESTS:

        We check that we correctly process results that depend on 32
        vs 64 bit architecture::

            sage: import sys
            sage: bitness = '64' if sys.maxsize > (1 << 32) else '32'
            sage: sys.maxsize == 2^63 - 1
            False # 32-bit
            True  # 64-bit
            sage: ex = doctests[20].examples[11]
            sage: ((bitness == '64' and ex.want == 'True  \n')
            ....:  or (bitness == '32' and ex.want == 'False \n'))
            True

        We check that lines starting with a # aren't doctested::

            #sage: raise RuntimeError
        """
        if not os.path.exists(self.path):
            import errno
            raise OSError(errno.ENOENT, "File does not exist", self.path)
        base, filename = os.path.split(self.path)
        _, ext = os.path.splitext(filename)
        if not self.in_lib and ext in ('.py', '.pyx', '.sage', '.spyx'):
            cwd = os.getcwd()
            if base:
                os.chdir(base)
            try:
                load(filename, namespace) # errors raised here will be caught in DocTestTask
            finally:
                os.chdir(cwd)
        self.qualified_name = NestedName(self.basename)
        return self._create_doctests(namespace)

    def _test_enough_doctests(self, check_extras=True, verbose=True):
        r"""
        This function checks to see that the doctests are not getting
        unexpectedly skipped.

        It uses a different (and simpler) code path than the doctest
        creation functions. In particular, it does not understand
        file-level and block-level # optional / needs tags.

        INPUT:

        - ``check_extras`` -- boolean (default: ``True``); whether to check if
          doctests are created that do not correspond to either a ``sage:``
          or a ``>>>`` prompt

        - ``verbose`` -- boolean (default: ``True``); whether to print
          offending line numbers when there are missing or extra tests

        TESTS::

            sage: # not tested (because the output will change when source files are changed)
            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: cwd = os.getcwd()
            sage: os.chdir(SAGE_SRC)
            sage: import itertools
            sage: for path, dirs, files in itertools.chain(os.walk('sage'), os.walk('doc')):
            ....:     path = os.path.relpath(path)
            ....:     dirs.sort(); files.sort()
            ....:     for F in files:
            ....:         _, ext = os.path.splitext(F)
            ....:         if ext in ('.py', '.pyx', '.pxd', '.pxi', '.sage', '.spyx', '.rst'):
            ....:             filename = os.path.join(path, F)
            ....:             FDS = FileDocTestSource(filename, DocTestDefaults(long=True, optional=True, force_lib=True))
            ....:             FDS._test_enough_doctests(verbose=False)
            There are 4 unexpected tests being run in sage/doctest/parsing.py
            There are 1 unexpected tests being run in sage/doctest/reporting.py
            sage: os.chdir(cwd)
        """
        expected = []
        rest = isinstance(self, RestSource)
        if rest:
            skipping = False
            in_block = False
            last_line = ''
        for lineno, line in self:
            if not line.strip():
                continue
            if rest:
                if line.strip().startswith(".. nodoctest"):
                    return
                # We need to track blocks in order to figure out whether we're skipping.
                if in_block:
                    indent = whitespace.match(line).end()
                    if indent <= starting_indent:
                        in_block = False
                        skipping = False
                if not in_block:
                    m1 = double_colon.match(line)
                    m2 = code_block.match(line.lower())
                    starting = (m1 and not line.strip().startswith(".. ")) or m2
                    if starting:
                        if ".. skip" in last_line:
                            skipping = True
                        in_block = True
                        starting_indent = whitespace.match(line).end()
                last_line = line
            if (not rest or in_block) and sagestart.match(line) and not ((rest and skipping) or untested.search(line.lower())):
                expected.append(lineno+1)
        actual = []
        tests, _ = self.create_doctests({})
        for dt in tests:
            if dt.examples:
                for ex in dt.examples[:-1]: # the last entry is a sig_on_count()
                    actual.append(dt.lineno + ex.lineno + 1)
        shortfall = sorted(set(expected).difference(set(actual)))
        extras = sorted(set(actual).difference(set(expected)))
        if len(actual) == len(expected):
            if not shortfall:
                return
            dif = extras[0] - shortfall[0]
            for e, s in zip(extras[1:],shortfall[1:]):
                if dif != e - s:
                    break
            else:
                print("There are %s tests in %s that are shifted by %s" % (len(shortfall), self.path, dif))
                if verbose:
                    print("    The correct line numbers are %s" % (", ".join(str(n) for n in shortfall)))
                return
        elif len(actual) < len(expected):
            print("There are %s tests in %s that are not being run" % (len(expected) - len(actual), self.path))
        elif check_extras:
            print("There are %s unexpected tests being run in %s" % (len(actual) - len(expected), self.path))
        if verbose:
            if shortfall:
                print("    Tests on lines %s are not run" % (", ".join(str(n) for n in shortfall)))
            if check_extras and extras:
                print("    Tests on lines %s seem extraneous" % (", ".join(str(n) for n in extras)))


class SourceLanguage:
    """
    An abstract class for functions that depend on the programming language of a doctest source.

    Currently supported languages include Python, ReST and LaTeX.
    """
    def parse_docstring(self, docstring, namespace, start):
        """
        Return a list of doctest defined in this docstring.

        This function is called by :meth:`DocTestSource._process_doc`.
        The default implementation, defined here, is to use the
        :class:`sage.doctest.parsing.SageDocTestParser` attached to
        this source to get doctests from the docstring.

        INPUT:

        - ``docstring`` -- string containing documentation and tests

        - ``namespace`` -- dictionary or :class:`sage.doctest.util.RecordingDict`

        - ``start`` -- integer; one less than the starting line number

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: from sage.doctest.parsing import SageDocTestParser
            sage: from sage.doctest.util import NestedName
            sage: filename = sage.doctest.util.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: doctests, _ = FDS.create_doctests({})
            sage: for dt in doctests:
            ....:     FDS.qualified_name = dt.name
            ....:     dt.examples = dt.examples[:-1] # strip off the sig_on() test
            ....:     assert(FDS.parse_docstring(dt.docstring,{},dt.lineno-1)[0] == dt)
        """
        return [self.parser.get_doctest(docstring, namespace, str(self.qualified_name),
                                        self.printpath, start + 1)]


class PythonSource(SourceLanguage):
    """
    This class defines the functions needed for the extraction of doctests from python sources.

    EXAMPLES::

        sage: from sage.doctest.control import DocTestDefaults
        sage: from sage.doctest.sources import FileDocTestSource
        sage: filename = sage.doctest.sources.__file__
        sage: FDS = FileDocTestSource(filename, DocTestDefaults())
        sage: type(FDS)
        <class 'sage.doctest.sources.PythonFileSource'>
    """
    # The same line can't both start and end a docstring
    start_finish_can_overlap = False

    def _init(self):
        """
        This function is called before creating doctests from a Python source.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.last_indent
            -1
        """
        self.last_indent = -1
        self.last_line = None
        self.quotetype = None
        self.paren_count = 0
        self.bracket_count = 0
        self.curly_count = 0
        self.code_wrapping = False

    def _update_quotetype(self, line):
        r"""
        Update the track of what kind of quoted string we are in.

        We need to track whether we're inside a triple quoted
        string, since a triple quoted string that starts a line
        could be the end of a string and thus not the beginning of a
        doctest (see sage.misc.sageinspect for an example).

        To do this tracking we need to track whether we're inside a
        string at all, since ''' inside a string doesn't start a
        triple quote (see the top of this file for an example).

        We also need to track parentheses and brackets, since we only
        want to update our record of last line and indentation level
        when the line is actually over.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS._update_quotetype('\"\"\"'); print(" ".join(list(FDS.quotetype)))
            " " "
            sage: FDS._update_quotetype("'''"); print(" ".join(list(FDS.quotetype)))
            " " "
            sage: FDS._update_quotetype('\"\"\"'); print(FDS.quotetype)
            None
            sage: FDS._update_quotetype("triple_quotes = re.compile(\"\\s*[rRuU]*((''')|(\\\"\\\"\\\"))\")")
            sage: print(FDS.quotetype)
            None
            sage: FDS._update_quotetype("''' Single line triple quoted string \\''''")
            sage: print(FDS.quotetype)
            None
            sage: FDS._update_quotetype("' Lots of \\\\\\\\'")
            sage: print(FDS.quotetype)
            None
        """
        def _update_parens(start, end=None):
            self.paren_count += line.count("(",start,end) - line.count(")",start,end)
            self.bracket_count += line.count("[",start,end) - line.count("]",start,end)
            self.curly_count += line.count("{",start,end) - line.count("}",start,end)
        pos = 0
        while pos < len(line):
            if self.quotetype is None:
                next_single = line.find("'",pos)
                next_double = line.find('"',pos)
                if next_single == -1 and next_double == -1:
                    next_comment = line.find("#",pos)
                    if next_comment == -1:
                        _update_parens(pos)
                    else:
                        _update_parens(pos,next_comment)
                    break
                elif next_single == -1:
                    m = next_double
                elif next_double == -1:
                    m = next_single
                else:
                    m = min(next_single, next_double)
                next_comment = line.find('#',pos,m)
                if next_comment != -1:
                    _update_parens(pos,next_comment)
                    break
                _update_parens(pos,m)
                if m+2 < len(line) and line[m] == line[m+1] == line[m+2]:
                    self.quotetype = line[m:m+3]
                    pos = m+3
                else:
                    self.quotetype = line[m]
                    pos = m+1
            else:
                next = line.find(self.quotetype,pos)
                if next == -1:
                    break
                elif next == 0 or line[next-1] != '\\':
                    pos = next + len(self.quotetype)
                    self.quotetype = None
                else:
                    # We need to worry about the possibility that
                    # there are an even number of backslashes before
                    # the quote, in which case it is not escaped
                    count = 1
                    slashpos = next - 2
                    while slashpos >= pos and line[slashpos] == '\\':
                        count += 1
                        slashpos -= 1
                    if count % 2 == 0:
                        pos = next + len(self.quotetype)
                        self.quotetype = None
                    else:
                        # The possible ending quote was escaped.
                        pos = next + 1

    def starting_docstring(self, line):
        """
        Determines whether the input line starts a docstring.

        If the input line does start a docstring (a triple quote),
        then this function updates ``self.qualified_name``.

        INPUT:

        - ``line`` -- string; one line of an input file

        OUTPUT: either ``None`` or a Match object

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: from sage.doctest.util import NestedName
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.starting_docstring("r'''")
            <...Match object...>
            sage: FDS.ending_docstring("'''")
            <...Match object...>
            sage: FDS.qualified_name = NestedName(FDS.basename)
            sage: FDS.starting_docstring("class MyClass():")
            sage: FDS.starting_docstring("    def hello_world(self):")
            sage: FDS.starting_docstring("        '''")
            <...Match object...>
            sage: FDS.qualified_name
            sage.doctest.sources.MyClass.hello_world
            sage: FDS.ending_docstring("    '''")
            <...Match object...>
            sage: FDS.starting_docstring("class NewClass():")
            sage: FDS.starting_docstring("    '''")
            <...Match object...>
            sage: FDS.ending_docstring("    '''")
            <...Match object...>
            sage: FDS.qualified_name
            sage.doctest.sources.NewClass
            sage: FDS.starting_docstring("print(")
            sage: FDS.starting_docstring("    '''Not a docstring")
            sage: FDS.starting_docstring("    ''')")
            sage: FDS.starting_docstring("def foo():")
            sage: FDS.starting_docstring("    '''This is a docstring'''")
            <...Match object...>
        """
        indent = whitespace.match(line).end()
        quotematch = None
        if self.quotetype is None and not self.code_wrapping:
            # We're not inside a triple quote and not inside code like
            # print(
            #     """Not a docstring
            #     """)

            if line[indent] != '#' and (indent == 0 or indent > self.last_indent):
                quotematch = triple_quotes.match(line)
                # It would be nice to only run the name_regex when
                # quotematch wasn't None, but then we mishandle classes
                # that don't have a docstring.
                if not self.code_wrapping and self.last_indent >= 0 and indent > self.last_indent:
                    name = name_regex.match(self.last_line)
                    if name:
                        name = name.groups()[0]
                        self.qualified_name[indent] = name
                    elif quotematch:
                        self.qualified_name[indent] = '?'
        self._update_quotetype(line)
        if line[indent] != '#' and not self.code_wrapping:
            self.last_line, self.last_indent = line, indent
        self.code_wrapping = not (self.paren_count == self.bracket_count == self.curly_count == 0)
        return quotematch

    def ending_docstring(self, line):
        r"""
        Determines whether the input line ends a docstring.

        INPUT:

        - ``line`` -- string, one line of an input file

        OUTPUT: an object that, when evaluated in a boolean context, gives
        ``True`` or ``False`` depending on whether the input line marks the
        end of a docstring

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: from sage.doctest.util import NestedName
            sage: filename = sage.doctest.sources.__file__
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.quotetype = "'''"
            sage: FDS.ending_docstring("'''")
            <...Match object...>
            sage: FDS.ending_docstring('\"\"\"')
        """
        quotematch = triple_quotes.match(line)
        if quotematch is not None and quotematch.groups()[0] != self.quotetype:
            quotematch = None
        self._update_quotetype(line)
        return quotematch

    def _neutralize_doctests(self, reindent):
        r"""
        Return a string containing the source of ``self``, but with
        doctests modified so they are not tested.

        This function is used in creating doctests for ReST files,
        since docstrings of Python functions defined inside verbatim
        blocks screw up Python's doctest parsing.

        INPUT:

        - ``reindent`` -- integer; the number of spaces to indent
          the result

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import StringDocTestSource, PythonSource
            sage: from sage.structure.dynamic_class import dynamic_class
            sage: s = "'''\n    sage: 2 + 2\n    4\n'''"
            sage: PythonStringSource = dynamic_class('PythonStringSource',
            ....:                                    (StringDocTestSource, PythonSource))
            sage: PSS = PythonStringSource('<runtime>', s, DocTestDefaults(), 'runtime')
            sage: print(PSS._neutralize_doctests(0))
            '''
                safe: 2 + 2
                4
            '''
        """
        neutralized = []
        in_docstring = False
        self._init()
        for lineno, line in self:
            if not line.strip():
                neutralized.append(line)
            elif in_docstring:
                if self.ending_docstring(line):
                    in_docstring = False
                neutralized.append(" "*reindent + find_prompt.sub(r"\1safe:\3",line))
            else:
                if self.starting_docstring(line):
                    in_docstring = True
                neutralized.append(" "*reindent + line)
        return "".join(neutralized)


class TexSource(SourceLanguage):
    """
    This class defines the functions needed for the extraction of
    doctests from a LaTeX source.

    EXAMPLES::

        sage: from sage.doctest.control import DocTestDefaults
        sage: from sage.doctest.sources import FileDocTestSource
        sage: filename = "sage_paper.tex"
        sage: FDS = FileDocTestSource(filename, DocTestDefaults())
        sage: type(FDS)
        <class 'sage.doctest.sources.TexFileSource'>
    """
    # The same line can't both start and end a docstring
    start_finish_can_overlap = False

    def _init(self):
        """
        This function is called before creating doctests from a Tex file.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = "sage_paper.tex"
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.skipping
            False
        """
        self.skipping = False

    def starting_docstring(self, line):
        r"""
        Determines whether the input line starts a docstring.

        Docstring blocks in tex files are defined by verbatim or
        lstlisting environments, and can be linked together by adding
        %link immediately after the \end{verbatim} or \end{lstlisting}.

        Within a verbatim (or lstlisting) block, you can tell Sage not to
        process the rest of the block by including a %skip line.

        INPUT:

        - ``line`` -- string, one line of an input file

        OUTPUT: boolean; whether the input line marks the start of a docstring
        (verbatim block)

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = "sage_paper.tex"
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()

        We start docstrings with \begin{verbatim} or \begin{lstlisting}::

            sage: FDS.starting_docstring(r"\begin{verbatim}")
            True
            sage: FDS.starting_docstring(r"\begin{lstlisting}")
            True
            sage: FDS.skipping
            False
            sage: FDS.ending_docstring("sage: 2+2")
            False
            sage: FDS.ending_docstring("4")
            False

        To start ignoring the rest of the verbatim block, use %skip::

            sage: FDS.ending_docstring("%skip")
            True
            sage: FDS.skipping
            True
            sage: FDS.starting_docstring("sage: raise RuntimeError")
            False

        You can even pretend to start another verbatim block while skipping::

            sage: FDS.starting_docstring(r"\begin{verbatim}")
            False
            sage: FDS.skipping
            True

        To stop skipping end the verbatim block::

            sage: FDS.starting_docstring(r"\end{verbatim} %link")
            False
            sage: FDS.skipping
            False

        Linking works even when the block was ended while skipping::

            sage: FDS.linking
            True
            sage: FDS.starting_docstring(r"\begin{verbatim}")
            True
        """
        if self.skipping:
            if self.ending_docstring(line, check_skip=False):
                self.skipping = False
            return False
        return bool(begin_verb.match(line) or begin_lstli.match(line))

    def ending_docstring(self, line, check_skip=True):
        r"""
        Determines whether the input line ends a docstring.

        Docstring blocks in tex files are defined by verbatim or
        lstlisting environments, and can be linked together by adding
        %link immediately after the \end{verbatim} or \end{lstlisting}.

        Within a verbatim (or lstlisting) block, you can tell Sage not to
        process the rest of the block by including a %skip line.

        INPUT:

        - ``line`` -- string, one line of an input file

        - ``check_skip`` -- boolean (default: ``True``); used internally in
          ``starting_docstring``

        OUTPUT: boolean; whether the input line marks the end of a docstring
        (verbatim block)

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = "sage_paper.tex"
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.ending_docstring(r"\end{verbatim}")
            True
            sage: FDS.ending_docstring(r"\end{lstlisting}")
            True
            sage: FDS.linking
            False

        Use %link to link with the next verbatim block::

            sage: FDS.ending_docstring(r"\end{verbatim}%link")
            True
            sage: FDS.linking
            True

        %skip also ends a docstring block::

            sage: FDS.ending_docstring("%skip")
            True
        """
        m = end_verb.match(line)
        if m:
            if m.groups()[0]:
                self.linking = True
            else:
                self.linking = False
            return True
        m = end_lstli.match(line)
        if m:
            if m.groups()[0]:
                self.linking = True
            else:
                self.linking = False
            return True
        if check_skip and skip.match(line):
            self.skipping = True
            return True
        return False


class RestSource(SourceLanguage):
    """
    This class defines the functions needed for the extraction of
    doctests from ReST sources.

    EXAMPLES::

        sage: from sage.doctest.control import DocTestDefaults
        sage: from sage.doctest.sources import FileDocTestSource
        sage: filename = "sage_doc.rst"
        sage: FDS = FileDocTestSource(filename, DocTestDefaults())
        sage: type(FDS)
        <class 'sage.doctest.sources.RestFileSource'>
    """
    # The same line can both start and end a docstring
    start_finish_can_overlap = True

    def _init(self):
        """
        This function is called before creating doctests from a ReST file.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = "sage_doc.rst"
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.link_all
            False
        """
        self.link_all = False
        self.last_line = ""
        self.last_indent = -1
        self.first_line = False
        self.skipping = False

    def starting_docstring(self, line):
        """
        A line ending with a double colon starts a verbatim block in a ReST file,
        as does a line containing ``.. CODE-BLOCK:: language``.

        This function also determines whether the docstring block
        should be joined with the previous one, or should be skipped.

        INPUT:

        - ``line`` -- string; one line of an input file

        OUTPUT: either ``None`` or a Match object

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = "sage_doc.rst"
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.starting_docstring("Hello world::")
            True
            sage: FDS.ending_docstring("    sage: 2 + 2")
            False
            sage: FDS.ending_docstring("    4")
            False
            sage: FDS.ending_docstring("We are now done")
            True
            sage: FDS.starting_docstring(".. link")
            sage: FDS.starting_docstring("::")
            True
            sage: FDS.linking
            True
        """
        if link_all.match(line):
            self.link_all = True
        if self.skipping:
            end_block = self.ending_docstring(line)
            if end_block:
                self.skipping = False
            else:
                return False
        m1 = double_colon.match(line)
        m2 = code_block.match(line.lower())
        starting = (m1 and not line.strip().startswith(".. ")) or m2
        if starting:
            self.linking = self.link_all or '.. link' in self.last_line
            self.first_line = True
            m = m1 or m2
            indent = len(m.groups()[0])
            if '.. skip' in self.last_line:
                self.skipping = True
                starting = False
        else:
            indent = self.last_indent
        self.last_line, self.last_indent = line, indent
        return starting

    def ending_docstring(self, line):
        """
        When the indentation level drops below the initial level the
        block ends.

        INPUT:

        - ``line`` -- string; one line of an input file

        OUTPUT: boolean; whether the verbatim block is ending

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: filename = "sage_doc.rst"
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS._init()
            sage: FDS.starting_docstring("Hello world::")
            True
            sage: FDS.ending_docstring("    sage: 2 + 2")
            False
            sage: FDS.ending_docstring("    4")
            False
            sage: FDS.ending_docstring("We are now done")
            True
        """
        if not line.strip():
            return False
        indent = whitespace.match(line).end()
        if self.first_line:
            self.first_line = False
            if indent <= self.last_indent:
                # We didn't indent at all
                return True
            self.last_indent = indent
        return indent < self.last_indent

    def parse_docstring(self, docstring, namespace, start):
        r"""
        Return a list of doctest defined in this docstring.

        Code blocks in a REST file can contain python functions with
        their own docstrings in addition to in-line doctests.  We want
        to include the tests from these inner docstrings, but Python's
        doctesting module has a problem if we just pass on the whole
        block, since it expects to get just a docstring, not the
        Python code as well.

        Our solution is to create a new doctest source from this code
        block and append the doctests created from that source.  We
        then replace the occurrences of "sage:" and ">>>" occurring
        inside a triple quote with "safe:" so that the doctest module
        doesn't treat them as tests.

        EXAMPLES::

            sage: from sage.doctest.control import DocTestDefaults
            sage: from sage.doctest.sources import FileDocTestSource
            sage: from sage.doctest.parsing import SageDocTestParser
            sage: from sage.doctest.util import NestedName
            sage: filename = "sage_doc.rst"
            sage: FDS = FileDocTestSource(filename, DocTestDefaults())
            sage: FDS.parser = SageDocTestParser(set(['sage']))
            sage: FDS.qualified_name = NestedName('sage_doc')
            sage: s = "Some text::\n\n    def example_python_function(a, \
            ....:      b):\n        '''\n        Brief description \
            ....:      of function.\n\n        EXAMPLES::\n\n            \
            ....:      sage: test1()\n            sage: test2()\n        \
            ....:      '''\n        return a + b\n\n    sage: test3()\n\nMore \
            ....:      ReST documentation."
            sage: tests = FDS.parse_docstring(s, {}, 100)
            sage: len(tests)
            2
            sage: for ex in tests[0].examples:
            ....:     print(ex.sage_source)
            test3()
            sage: for ex in tests[1].examples:
            ....:     print(ex.sage_source)
            test1()
            test2()
            sig_on_count() # check sig_on/off pairings (virtual doctest)
        """
        PythonStringSource = dynamic_class("sage.doctest.sources.PythonStringSource",
                                           (StringDocTestSource, PythonSource))
        min_indent = self.parser._min_indent(docstring)
        pysource = '\n'.join(l[min_indent:] for l in docstring.split('\n'))
        inner_source = PythonStringSource(self.basename, pysource,
                                          self.options,
                                          self.printpath,
                                          lineno_shift=start + 1)
        inner_doctests, _ = inner_source._create_doctests(namespace, True)
        safe_docstring = inner_source._neutralize_doctests(min_indent)
        outer_doctest = self.parser.get_doctest(safe_docstring, namespace,
                                                str(self.qualified_name),
                                                self.printpath, start + 1)
        return [outer_doctest] + inner_doctests


class DictAsObject(dict):
    """
    A simple subclass of dict that inserts the items from the initializing dictionary into attributes.

    EXAMPLES::

        sage: from sage.doctest.sources import DictAsObject
        sage: D = DictAsObject({'a':2})
        sage: D.a
        2
    """
    def __init__(self, attrs):
        """
        Initialization.

        INPUT:

        - ``attrs`` -- dictionary

        EXAMPLES::

            sage: from sage.doctest.sources import DictAsObject
            sage: D = DictAsObject({'a':2})
            sage: D.a == D['a']
            True
            sage: D.a
            2
        """
        super().__init__(attrs)
        self.__dict__.update(attrs)

    def __setitem__(self, ky, val):
        """
        We preserve the ability to access entries through either the
        dictionary or attribute interfaces.

        EXAMPLES::

            sage: from sage.doctest.sources import DictAsObject
            sage: D = DictAsObject({})
            sage: D['a'] = 2
            sage: D.a
            2
        """
        super().__setitem__(ky, val)
        try:
            super().__setattr__(ky, val)
        except TypeError:
            pass

    def __setattr__(self, ky, val):
        """
        We preserve the ability to access entries through either the
        dictionary or attribute interfaces.

        EXAMPLES::

            sage: from sage.doctest.sources import DictAsObject
            sage: D = DictAsObject({})
            sage: D.a = 2
            sage: D['a']
            2
        """
        super().__setitem__(ky, val)
        super().__setattr__(ky, val)



# sage_setup: distribution = sagemath-repl
"""
Utility functions

This module contains various utility functions and classes used in doctesting.

AUTHORS:

- David Roe (2012-03-27) -- initial version, based on Robert Bradshaw's code.
"""

# ****************************************************************************
#       Copyright (C) 2012 David Roe <roed.math@gmail.com>
#                     2012 Robert Bradshaw <robertwb@gmail.com>
#                     2012 William Stein <wstein@gmail.com>
#                     2013 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#                     2014 Volker Braun
#                     2017 Frédéric Chapoton
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from time import time as walltime
from os import sysconf, times
from contextlib import contextmanager
from cysignals.alarm import alarm, cancel_alarm, AlarmInterrupt


def count_noun(number, noun, plural=None, pad_number=False, pad_noun=False):
    """
    EXAMPLES::

        sage: from sage.doctest.util import count_noun
        sage: count_noun(1, "apple")
        '1 apple'
        sage: count_noun(1, "apple", pad_noun=True)
        '1 apple '
        sage: count_noun(1, "apple", pad_number=3)
        '  1 apple'
        sage: count_noun(2, "orange")
        '2 oranges'
        sage: count_noun(3, "peach", "peaches")
        '3 peaches'
        sage: count_noun(1, "peach", plural='peaches', pad_noun=True)
        '1 peach  '
    """
    if plural is None:
        plural = noun + "s"
    if pad_noun:
        # We assume that the plural is never shorter than the noun....
        pad_noun = " " * (len(plural) - len(noun))
    else:
        pad_noun = ""
    if pad_number:
        number_str = ("%%%sd" % pad_number) % number
    else:
        number_str = "%d" % number
    if number == 1:
        return "%s %s%s" % (number_str, noun, pad_noun)
    else:
        return "%s %s" % (number_str, plural)


def dict_difference(self, other):
    """
    Return a dict with all key-value pairs occurring in ``self`` but not
    in ``other``.

    EXAMPLES::

        sage: from sage.doctest.util import dict_difference
        sage: d1 = {1: 'a', 2: 'b', 3: 'c'}
        sage: d2 = {1: 'a', 2: 'x', 4: 'c'}
        sage: dict_difference(d2, d1)
        {2: 'x', 4: 'c'}

    ::

        sage: from sage.doctest.control import DocTestDefaults
        sage: D1 = DocTestDefaults()
        sage: D2 = DocTestDefaults(foobar='hello', timeout=100)
        sage: dict_difference(D2.__dict__, D1.__dict__)
        {'foobar': 'hello', 'timeout': 100}
    """
    D = {}
    for k, v in self.items():
        try:
            if other[k] == v:
                continue
        except KeyError:
            pass
        D[k] = v
    return D


class Timer:
    """
    A simple timer.

    EXAMPLES::

        sage: from sage.doctest.util import Timer
        sage: Timer()
        {}
        sage: TestSuite(Timer()).run()
    """

    def _proc_stat_cpu_seconds(self, path):
        r"""
        Parse a "stat" file from the ``/proc`` filesystem to get
        the cputime of a process.

        This also includes the times for child processes, but only
        those that have already terminated and for which ``wait()``
        was called. It is important to note that pexpect processes DO
        NOT fall into that category.

        The document ``Documentation/filesystems/proc.rst`` within the
        Linux kernel source tree defines a "stat" file.

        INPUT:

        - ``path`` -- string; the path to a "stat" file on the ``/proc``
          filesystem, typically "/proc/<pid>/stat", from which we will
          read cputime information

        OUTPUT:

        A nonnegative float representing the number of cpu-seconds
        used by the process associated with ``path``. An ``OSError`` is
        raised if anything goes wrong, which typically happens on
        platforms that don't store this information under ``/proc``.

        TESTS:

        About all we can say for certain is that this will return a
        nonnegative float or raise an ``OSError``::

            sage: from sage.doctest.util import Timer
            sage: cputime = float(0.0)
            sage: path = "/proc/1/stat"
            sage: try:
            ....:     cputime = Timer()._proc_stat_cpu_seconds(path)
            ....: except OSError:
            ....:     pass
            sage: cputime >= 0.0
            True
            sage: isinstance(cputime, float)
            True

        We can force an ``OSError`` with an invalid PID::

            sage: from sage.doctest.util import Timer
            sage: path = "/proc/-1/stat"
            sage: cputime = Timer()._proc_stat_cpu_seconds(path)
            Traceback (most recent call last):
            ...
            OSError: unable to access /proc/-1/stat

        Or with an unparseable file (wrong number of fields, non-float
        fields, et cetera)::

            sage: from tempfile import NamedTemporaryFile
            sage: from os import unlink
            sage: from sage.doctest.util import Timer
            sage: with NamedTemporaryFile(delete=False, mode="w") as f:
            ....:     _ = f.write("1 2 3 4 5")
            sage: cputime = Timer()._proc_stat_cpu_seconds(f.name)
            Traceback (most recent call last):
            ...
            OSError: unable to parse ...
            sage: os.unlink(f.name)
            sage: with NamedTemporaryFile(delete=False, mode="w") as f:
            ....:     _ = f.write("1 2 3 4 5 6 7 8 9 10 11 12 w x y z 17")
            sage: cputime = Timer()._proc_stat_cpu_seconds(f.name)
            Traceback (most recent call last):
            ...
            OSError: unable to parse ...
            sage: os.unlink(f.name)

        """
        try:
            with open(path, "r") as statfile:
                stats = statfile.read().split()
        except (FileNotFoundError, PermissionError) as e:
            # FileNotFoundError: bad PID, or no /proc support
            # PermissionError: can't read the stat file
            raise OSError(f"unable to access {path}") from e

        if len(stats) < 17:
            raise OSError(f"unable to parse {path}")

        try:
            # These fields used to be documented in the proc(5) man
            # page, but are now most easily found in the Linux kernel
            # documentation (Documentation/filesystems/proc.rst). The
            # intent is to sum the user- and kernel-mode "jiffies" for
            # both the given process and its children.
            cputicks = sum( float(s) for s in stats[13:17] )
        except (ArithmeticError, TypeError, ValueError) as e:
            # ArithmeticError: unexpected (non-numeric?) values in fields
            # TypeError/ValueError: fields can't be converted to float
            raise OSError(f"unable to parse {path}") from e

        try:
            hertz = sysconf("SC_CLK_TCK")
        except (ValueError) as e:
            # ValueError: SC_CLK_TCK doesn't exist
            raise OSError("SC_CLK_TCK sysconf not found") from e

        if hertz <= 0:
            # The python documentation for os.sysconf() says, "If the
            # configuration value specified by name isn’t defined, -1
            # is returned." Having tried this with a junk value, I
            # don't believe it: I got a ValueError that was handled
            # above. Nevertheless, we play it safe here and turn a -1
            # into an OSError. We check for zero, too, because we're
            # about to divide by it.
            raise OSError("SC_CLK_TCK sysconf is nonpositive")

        return (cputicks / hertz)

    def _quick_cputime(self, expect_objects):
        r"""
        A fast replacement for ``sage.misc.timing.cputime``.

        This is a "reliable" replacement (on Linux/BSD) that takes
        subprocesses (particularly pexpect interfaces) into
        account. The ``cputime()`` function from the ``misc`` module
        can be passed ``subprocesses=True``, but this has a few
        faults; mainly that it relies on each pexpect interface to
        implement its own ``cputime()`` function. And most of our
        pexpect interfaces either don't implement one, or implement
        one in a way that requires the subprocess (being pexpected) to
        be in perfect working condition -- that will often not be the
        case at the end of a doctest line.

        INPUT:

        - ``expect_objects`` -- list; a list of
          :class:`sage.interfaces.expect.Expect` instances whose CPU
          times will be included in the total

        OUTPUT:

        A float measuring the cputime in seconds of the sage process
        and all its subprocesses.

        TESTS:

        About all we can say for certain is that this will return a
        nonnegative float::

            sage: from sage.doctest.util import Timer
            sage: from sage.interfaces.quit import expect_objects
            sage: cputime = Timer()._quick_cputime(expect_objects)
            sage: cputime >= 0.0
            True
            sage: isinstance(cputime, float)
            True

        If an error occurs in :meth:`_proc_stat_cpu_seconds`, this
        function should still return a valid answer, albeit one that
        is missing timing information for the PID that failed::

            sage: class FakeExpect:
            ....:     def __call__(self):
            ....:         return self
            ....:     def is_running(self):
            ....:         return True
            ....:     def pid(self):
            ....:         return -1
            sage: e = FakeExpect()
            sage: from sage.doctest.util import Timer
            sage: cputime = Timer()._quick_cputime([e])
            sage: cputime >= 0.0
            True
            sage: isinstance(cputime, float)
            True
        """
        # Start by using os.times() to get the cputime for sage itself
        # and any subprocesses that have been wait()ed for and that
        # have terminated.
        cputime = sum( times()[:4] )

        # Now try to get the times for any pexpect interfaces, since
        # they do not fall into the category above.
        for s in expect_objects:
            S = s()
            if S and S.is_running():
                try:
                    # This will fail anywhere but linux/BSD, but
                    # there's no good cross-platform way to get the
                    # cputimes from pexpect interfaces without totally
                    # mucking up the doctests.
                    path = f"/proc/{S.pid()}/stat"
                    cputime += self._proc_stat_cpu_seconds(path)
                except OSError:
                    # If we're on macOS, we can fall back to using
                    # psutil, but only if it's installed. It's usually
                    # installed as a transitive dependency (ipython
                    # needs it), but it isn't explicitly listed as
                    # a dependency of sagelib.
                    try:
                        from psutil import (NoSuchProcess,
                                            Process,
                                            ZombieProcess)
                        try:
                            cputime += sum(Process(S.pid()).cpu_times()[0:2])
                        except (ValueError, NoSuchProcess, ZombieProcess):
                            # ValueError: invalid (e.g. negative) PID
                            # NoSuchProcess: it's gone
                            # ZombieProcess: PID refers to a zombie
                            pass
                    except ImportError:
                        pass

        return cputime

    def start(self):
        """
        Start the timer.

        Can be called multiple times to reset the timer.

        EXAMPLES::

            sage: from sage.doctest.util import Timer
            sage: Timer().start()
            {'cputime': ..., 'walltime': ...}
        """
        from sage.interfaces.quit import expect_objects
        self.cputime = self._quick_cputime(expect_objects)
        self.walltime = walltime()
        return self

    def stop(self):
        """
        Stops the timer, recording the time that has passed since it
        was started.

        EXAMPLES::

            sage: from sage.doctest.util import Timer
            sage: import time
            sage: timer = Timer().start()
            sage: time.sleep(float(0.5))
            sage: timer.stop()
            {'cputime': ..., 'walltime': ...}
        """
        from sage.interfaces.quit import expect_objects
        self.cputime = self._quick_cputime(expect_objects) - self.cputime
        self.walltime = walltime() - self.walltime
        return self

    def annotate(self, object):
        """
        Annotates the given object with the cputime and walltime
        stored in this timer.

        EXAMPLES::

            sage: from sage.doctest.util import Timer
            sage: Timer().start().annotate(EllipticCurve)
            sage: EllipticCurve.cputime # random
            2.817255
            sage: EllipticCurve.walltime # random
            1332649288.410404
        """
        object.cputime = self.cputime
        object.walltime = self.walltime

    def __repr__(self):
        """
        String representation is from the dictionary.

        EXAMPLES::

            sage: from sage.doctest.util import Timer
            sage: repr(Timer().start()) # indirect doctest
            "{'cputime': ..., 'walltime': ...}"
        """
        return str(self)

    def __str__(self):
        """
        String representation is from the dictionary.

        EXAMPLES::

            sage: from sage.doctest.util import Timer
            sage: str(Timer().start()) # indirect doctest
            "{'cputime': ..., 'walltime': ...}"
        """
        return str(self.__dict__)

    def __eq__(self, other):
        """
        Comparison.

        EXAMPLES::

            sage: from sage.doctest.util import Timer
            sage: Timer() == Timer()
            True
            sage: t = Timer().start()
            sage: loads(dumps(t)) == t
            True
        """
        if not isinstance(other, Timer):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Test for non-equality.

        EXAMPLES::

            sage: from sage.doctest.util import Timer
            sage: Timer() == Timer()
            True
            sage: t = Timer().start()
            sage: loads(dumps(t)) != t
            False
        """
        return not (self == other)


# Inheritance rather then delegation as globals() must be a dict
class RecordingDict(dict):
    """
    This dictionary is used for tracking the dependencies of an example.

    This feature allows examples in different doctests to be grouped
    for better timing data.  It's obtained by recording whenever
    anything is set or retrieved from this dictionary.

    EXAMPLES::

        sage: from sage.doctest.util import RecordingDict
        sage: D = RecordingDict(test=17)
        sage: D.got
        set()
        sage: D['test']
        17
        sage: D.got
        {'test'}
        sage: D.set
        set()
        sage: D['a'] = 1
        sage: D['a']
        1
        sage: D.set
        {'a'}
        sage: D.got
        {'test'}

    TESTS::

        sage: TestSuite(D).run()
    """
    def __init__(self, *args, **kwds):
        """
        Initialization arguments are the same as for a normal dictionary.

        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: D.got
            set()
        """
        dict.__init__(self, *args, **kwds)
        self.start()

    def start(self):
        """
        We track which variables have been set or retrieved.
        This function initializes these lists to be empty.

        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: D.set
            set()
            sage: D['a'] = 4
            sage: D.set
            {'a'}
            sage: D.start(); D.set
            set()
        """
        self.set = set()
        self.got = set()

    def __getitem__(self, name):
        """
        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: D['a'] = 4
            sage: D.got
            set()
            sage: D['a'] # indirect doctest
            4
            sage: D.got
            set()
            sage: D['d']
            42
            sage: D.got
            {'d'}
        """
        if name not in self.set:
            self.got.add(name)
        return dict.__getitem__(self, name)

    def __setitem__(self, name, value):
        """
        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: D['a'] = 4 # indirect doctest
            sage: D.set
            {'a'}
        """
        self.set.add(name)
        dict.__setitem__(self, name, value)

    def __delitem__(self, name):
        """
        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: del D['d'] # indirect doctest
            sage: D.set
            {'d'}
        """
        self.set.add(name)
        dict.__delitem__(self, name)

    def get(self, name, default=None):
        """
        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: D.get('d')
            42
            sage: D.got
            {'d'}
            sage: D.get('not_here')
            sage: sorted(list(D.got))
            ['d', 'not_here']
        """
        if name not in self.set:
            self.got.add(name)
        return dict.get(self, name, default)

    def copy(self):
        """
        Note that set and got are not copied.

        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: D['a'] = 4
            sage: D.set
            {'a'}
            sage: E = D.copy()
            sage: E.set
            set()
            sage: sorted(E.keys())
            ['a', 'd']
        """
        return RecordingDict(dict.copy(self))

    def __reduce__(self):
        """
        Pickling.

        EXAMPLES::

            sage: from sage.doctest.util import RecordingDict
            sage: D = RecordingDict(d = 42)
            sage: D['a'] = 4
            sage: D.get('not_here')
            sage: E = loads(dumps(D))
            sage: E.got
            {'not_here'}
        """
        return make_recording_dict, (dict(self), self.set, self.got)


def make_recording_dict(D, st, gt):
    """
    Auxiliary function for pickling.

    EXAMPLES::

        sage: from sage.doctest.util import make_recording_dict
        sage: D = make_recording_dict({'a':4,'d':42},set(),set(['not_here']))
        sage: sorted(D.items())
        [('a', 4), ('d', 42)]
        sage: D.got
        {'not_here'}
    """
    ans = RecordingDict(D)
    ans.set = st
    ans.got = gt
    return ans


class NestedName:
    """
    Class used to construct fully qualified names based on indentation level.

    EXAMPLES::

        sage: from sage.doctest.util import NestedName
        sage: qname = NestedName('sage.categories.algebras')
        sage: qname[0] = 'Algebras'; qname
        sage.categories.algebras.Algebras
        sage: qname[4] = '__contains__'; qname
        sage.categories.algebras.Algebras.__contains__
        sage: qname[4] = 'ParentMethods'
        sage: qname[8] = 'from_base_ring'; qname
        sage.categories.algebras.Algebras.ParentMethods.from_base_ring

    TESTS::

        sage: TestSuite(qname).run()
    """
    def __init__(self, base):
        """
        INPUT:

        - ``base`` -- string; the name of the module

        EXAMPLES::

            sage: from sage.doctest.util import NestedName
            sage: qname = NestedName('sage.categories.algebras')
            sage: qname
            sage.categories.algebras
        """
        self.all = [base]

    def __setitem__(self, index, value):
        """
        Set the value at a given indentation level.

        INPUT:

        - ``index`` -- positive integer; the indentation level (often a
          multiple of 4, but not necessarily)
        - ``value`` -- string; the name of the class or function at that
          indentation level

        EXAMPLES::

            sage: from sage.doctest.util import NestedName
            sage: qname = NestedName('sage.categories.algebras')
            sage: qname[1] = 'Algebras' # indirect doctest
            sage: qname
            sage.categories.algebras.Algebras
            sage: qname.all
            ['sage.categories.algebras', None, 'Algebras']
        """
        if index < 0:
            raise ValueError
        while len(self.all) <= index:
            self.all.append(None)
        self.all[index+1:] = [value]

    def __str__(self):
        """
        Return a .-separated string giving the full name.

        EXAMPLES::

            sage: from sage.doctest.util import NestedName
            sage: qname = NestedName('sage.categories.algebras')
            sage: qname[1] = 'Algebras'
            sage: qname[44] = 'at_the_end_of_the_universe'
            sage: str(qname) # indirect doctest
            'sage.categories.algebras.Algebras.at_the_end_of_the_universe'
        """
        return repr(self)

    def __repr__(self):
        """
        Return a .-separated string giving the full name.

        EXAMPLES::

            sage: from sage.doctest.util import NestedName
            sage: qname = NestedName('sage.categories.algebras')
            sage: qname[1] = 'Algebras'
            sage: qname[44] = 'at_the_end_of_the_universe'
            sage: print(qname) # indirect doctest
            sage.categories.algebras.Algebras.at_the_end_of_the_universe
        """
        return '.'.join(a for a in self.all if a is not None)

    def __eq__(self, other):
        """
        Comparison is just comparison of the underlying lists.

        EXAMPLES::

            sage: from sage.doctest.util import NestedName
            sage: qname = NestedName('sage.categories.algebras')
            sage: qname2 = NestedName('sage.categories.algebras')
            sage: qname == qname2
            True
            sage: qname[0] = 'Algebras'
            sage: qname2[2] = 'Algebras'
            sage: repr(qname) == repr(qname2)
            True
            sage: qname == qname2
            False
        """
        if not isinstance(other, NestedName):
            return False
        return self.all == other.all

    def __ne__(self, other):
        """
        Test for non-equality.

        EXAMPLES::

            sage: from sage.doctest.util import NestedName
            sage: qname = NestedName('sage.categories.algebras')
            sage: qname2 = NestedName('sage.categories.algebras')
            sage: qname != qname2
            False
            sage: qname[0] = 'Algebras'
            sage: qname2[2] = 'Algebras'
            sage: repr(qname) == repr(qname2)
            True
            sage: qname != qname2
            True
        """
        return not (self == other)


@contextmanager
def ensure_interruptible_after(seconds: float, max_wait_after_interrupt: float = 0.2, inaccuracy_tolerance: float = 0.1):
    """
    Helper function for doctesting to ensure that the code is interruptible after a certain amount of time.
    This should only be used for internal doctesting purposes.

    EXAMPLES::

        sage: from sage.doctest.util import ensure_interruptible_after
        sage: with ensure_interruptible_after(1) as data: sleep(3)

    ``as data`` is optional, but if it is used, it will contain a few useful values::

        sage: data  # abs tol 1
        {'alarm_raised': True, 'elapsed': 1.0}

    ``max_wait_after_interrupt`` can be passed if the function may take longer than usual to be interrupted::

        sage: # needs sage.misc.cython
        sage: cython(r'''
        ....: from posix.time cimport clock_gettime, CLOCK_REALTIME, timespec, time_t
        ....: from cysignals.signals cimport sig_check
        ....:
        ....: cpdef void uninterruptible_sleep(double seconds):
        ....:     cdef timespec start_time, target_time
        ....:     clock_gettime(CLOCK_REALTIME, &start_time)
        ....:
        ....:     cdef time_t floor_seconds = <time_t>seconds
        ....:     target_time.tv_sec = start_time.tv_sec + floor_seconds
        ....:     target_time.tv_nsec = start_time.tv_nsec + <long>((seconds - floor_seconds) * 1e9)
        ....:     if target_time.tv_nsec >= 1000000000:
        ....:         target_time.tv_nsec -= 1000000000
        ....:         target_time.tv_sec += 1
        ....:
        ....:     while True:
        ....:         clock_gettime(CLOCK_REALTIME, &start_time)
        ....:         if start_time.tv_sec > target_time.tv_sec or (start_time.tv_sec == target_time.tv_sec and start_time.tv_nsec >= target_time.tv_nsec):
        ....:             break
        ....:
        ....: cpdef void check_interrupt_only_occasionally():
        ....:     for i in range(10):
        ....:         uninterruptible_sleep(0.8)
        ....:         sig_check()
        ....: ''')
        sage: with ensure_interruptible_after(1):  # not passing max_wait_after_interrupt will raise an error
        ....:     check_interrupt_only_occasionally()
        Traceback (most recent call last):
        ...
        RuntimeError: Function is not interruptible within 1.0000 seconds, only after 1.60... seconds
        sage: with ensure_interruptible_after(1, max_wait_after_interrupt=0.9):
        ....:     check_interrupt_only_occasionally()

    TESTS::

        sage: with ensure_interruptible_after(2) as data: sleep(1)
        Traceback (most recent call last):
        ...
        RuntimeError: Function terminates early after 1... < 2.0000 seconds
        sage: data  # abs tol 1
        {'alarm_raised': False, 'elapsed': 1.0}

    The test above requires a large tolerance, because both ``time.sleep`` and
    ``from posix.unistd cimport usleep`` may have slowdown on the order of 0.1s on Mac,
    likely because the system is idle and GitHub CI switches the program out,
    and context switch back takes time. Besides, there is an issue with ``Integer``
    destructor, see `<https://github.com/sagemath/cysignals/issues/215>`_
    So we use busy wait and Python integers::

        sage: # needs sage.misc.cython
        sage: cython(r'''
        ....: from posix.time cimport clock_gettime, CLOCK_REALTIME, timespec, time_t
        ....: from cysignals.signals cimport sig_check
        ....:
        ....: cpdef void interruptible_sleep(double seconds):
        ....:     cdef timespec start_time, target_time
        ....:     clock_gettime(CLOCK_REALTIME, &start_time)
        ....:
        ....:     cdef time_t floor_seconds = <time_t>seconds
        ....:     target_time.tv_sec = start_time.tv_sec + floor_seconds
        ....:     target_time.tv_nsec = start_time.tv_nsec + <long>((seconds - floor_seconds) * 1e9)
        ....:     if target_time.tv_nsec >= 1000000000:
        ....:         target_time.tv_nsec -= 1000000000
        ....:         target_time.tv_sec += 1
        ....:
        ....:     while True:
        ....:         sig_check()
        ....:         clock_gettime(CLOCK_REALTIME, &start_time)
        ....:         if start_time.tv_sec > target_time.tv_sec or (start_time.tv_sec == target_time.tv_sec and start_time.tv_nsec >= target_time.tv_nsec):
        ....:             break
        ....: ''')
        sage: with ensure_interruptible_after(2) as data: interruptible_sleep(1r)
        Traceback (most recent call last):
        ...
        RuntimeError: Function terminates early after 1.00... < 2.0000 seconds
        sage: with ensure_interruptible_after(1) as data: uninterruptible_sleep(2r)
        Traceback (most recent call last):
        ...
        RuntimeError: Function is not interruptible within 1.0000 seconds, only after 2.00... seconds
        sage: data  # abs tol 0.01
        {'alarm_raised': True, 'elapsed': 2.0}
        sage: with ensure_interruptible_after(1): uninterruptible_sleep(2r); raise RuntimeError
        Traceback (most recent call last):
        ...
        RuntimeError: Function is not interruptible within 1.0000 seconds, only after 2.00... seconds
        sage: data  # abs tol 0.01
        {'alarm_raised': True, 'elapsed': 2.0}

    ::

        sage: with ensure_interruptible_after(1) as data: raise ValueError
        Traceback (most recent call last):
        ...
        ValueError
        sage: data  # abs tol 0.01
        {'alarm_raised': False, 'elapsed': 0.0}
    """
    seconds = float(seconds)
    max_wait_after_interrupt = float(max_wait_after_interrupt)
    inaccuracy_tolerance = float(inaccuracy_tolerance)
    # use Python float to avoid slowdown with Sage Integer (see https://github.com/sagemath/cysignals/issues/215)
    data = {}
    start_time = walltime()
    alarm(seconds)
    alarm_raised = False

    try:
        yield data
    except AlarmInterrupt as e:
        e.__traceback__ = None  # workaround for https://github.com/python/cpython/pull/129276
        alarm_raised = True
    finally:
        before_cancel_alarm_elapsed = walltime() - start_time
        cancel_alarm()
        elapsed = walltime() - start_time
        data["elapsed"] = elapsed
        data["alarm_raised"] = alarm_raised

    if elapsed > seconds + max_wait_after_interrupt:
        raise RuntimeError(
                f"Function is not interruptible within {seconds:.4f} seconds, only after {elapsed:.4f} seconds"
                + ("" if alarm_raised else " (__exit__ called before interrupt check)"))

    if alarm_raised:
        if elapsed < seconds - inaccuracy_tolerance:
            raise RuntimeError(f"Interrupted too early: {elapsed:.4f} < {seconds:.4f}, this should not happen")
    else:
        raise RuntimeError(f"Function terminates early after {elapsed:.4f} < {seconds:.4f} seconds")
